from __future__ import annotations

from typing import Self, override

import jax.numpy as jnp
import jax.random as jr
from jax import tree

from .annotations import JaxArray


class Projectable:
    """Base class to override how an element of a PyTree is projected."""

    def project(self, projector: Projector) -> Self:
        raise NotImplementedError


class Projector:
    """A tool to project the arrays in a PyTree to fewer dimensions. This is useful for graphing."""

    @override
    def __init__(self, *, seed: int = 0, dimensions: int = 2) -> None:
        super().__init__()
        self.seed = seed
        self.dimensions = dimensions
        self._projection_matrices: dict[int, JaxArray] = {}

    def project[T](self, projectable: T) -> T:
        """Project the arrays in a PyTree to the plane."""
        return tree.map(self._project, projectable, is_leaf=lambda x: isinstance(x, Projectable))

    def get_projection_matrix(self, features: int) -> JaxArray:
        """Produce a projection matrix."""
        if features <= self.dimensions:
            return jnp.eye(features)
        if features not in self._projection_matrices:
            self._projection_matrices[features] = self._random_directions(features)
        return self._projection_matrices[features]

    def _project[T](self, projectable: T) -> T:
        """Project an array or Projectable instance to the plane."""
        if projectable is None:
            return projectable
        if isinstance(projectable, Projectable):
            return projectable.project(self)  # type: ignore

        assert isinstance(projectable, JaxArray)
        features = projectable.shape[-1]
        if features <= self.dimensions:
            return projectable
        projection_matrix = self.get_projection_matrix(features)
        return projectable @ projection_matrix  # type: ignore # pyright: ignore

    def _random_directions(self, features: int) -> JaxArray:
        """Produce a random projection matrix.

        Returns: A matrix of random numbers with `features` rows and `self.dimensions` columns, each
            column of which is mutually orthogonal and unit-length.
        """
        if features < self.dimensions:
            raise ValueError
        if self.seed == 0 and features == self.dimensions:
            return jnp.eye(features)
        key = jr.key(self.seed)
        a = jr.uniform(key, shape=(features, self.dimensions))
        result = jnp.linalg.qr(a)
        assert not isinstance(result, JaxArray)
        return result.Q



from __future__ import annotations

import math
import operator

import jax.numpy as jnp
from jax import tree

from .annotations import JaxArray, JaxBooleanArray


def dynamic_tree_all(x: object, /) -> JaxBooleanArray:
    """Like `jax.tree.all`, but can be used in dynamic code like jitted functions and loops."""
    jax_true = jnp.asarray(True)  # noqa: FBT003
    return tree.reduce_associative(jnp.logical_and, x, identity=jax_true)


def tree_sum(x: object, /) -> JaxArray:
    retval = tree.reduce_associative(operator.add, tree.map(jnp.sum, x), identity=jnp.zeros(()))
    assert isinstance(retval, JaxArray)
    assert retval.ndim == 0
    return retval


def element_count(x: object, /) -> int:
    def array_element_count(x: object) -> int:
        if not isinstance(x, JaxArray):
            raise TypeError
        return math.prod(x.shape)

    retval = tree.reduce_associative(operator.add, tree.map(array_element_count, x), identity=0)
    assert isinstance(retval, int)
    return retval



import jax.numpy as jnp
import pytest

from tjax import element_count, tree_sum


@pytest.mark.parametrize(
    ("x", "s"), [(jnp.arange(3, 10), 42), ((jnp.arange(5, 9), jnp.arange(4, 12)), 86), ((), 0)]
)
def test_tree_sum(x: object, s: float) -> None:
    t = tree_sum(x)
    assert t.ndim == 0
    assert float(t) == s


@pytest.mark.parametrize(
    ("x", "s"), [(jnp.arange(3, 10), 7), ((jnp.arange(5, 9), jnp.arange(4, 12)), 12), ((), 0)]
)
def test_element_count(x: object, s: int) -> None:
    ec = element_count(x)
    assert ec == s



from __future__ import annotations

from array_api_compat import array_namespace

from .annotations import Array, DType, Namespace


def cast_to_result_type[T: tuple[Array | None, ...]](
    arrays: T, *arrays_and_dtypes: type[complex | bool] | Array | DType, xp: Namespace | None = None
) -> T:
    """Casts the input arrays to a common data dtype.

    Args:
        arrays: Arrays to be converted (or None).
        arrays_and_dtypes: Extra arrays and types to be passed to result_type.
        xp: The optional namespace.

    Returns:
        The arrays cast to the inferred type.
    """
    if xp is None:
        xp = array_namespace(*arrays)
    dtype = xp.result_type(*arrays, *arrays_and_dtypes)
    return tuple(  # type: ignore
        xp.astype(x, dtype)  # pyright: ignore
        if x is not None
        else None
        for x in arrays
    )



from __future__ import annotations

from typing import SupportsInt

import jax
import jax.numpy as jnp
import jax.random as jr
import numpy as np
from jax import Array, debug, tree
from rich.console import Console

from tjax._src.dataclasses.dataclass import dataclass
from .internal import internal_print_generic


@dataclass
class JaxKey:
    key_data: Array


def replace_key(leaf: object) -> object:
    if isinstance(leaf, Array) and jnp.issubdtype(leaf.dtype, jax.dtypes.prng_key):
        return JaxKey(jr.key_data(leaf))
    return leaf


def print_generic(
    *args: object,
    immediate: bool = False,
    raise_on_nan: bool = True,
    as_leaves: bool = False,
    console: Console | None = None,
) -> None:
    """Uses internal_print_generic in a tapped function.

    Args:
        immediate: Print the value immediately even if they're tracers.  Do not wait for the
            compiled function to be called.
        raise_on_nan: Assert if NaN is found in the output.
        as_leaves: Print the elements of args showing only the the tree leaves and their paths.
        console: The console that formats the output.
        args: Positional arguments to be printed.  Only dynamic arguments are allowed.
    """
    args = tree.map(replace_key, args)

    if immediate:
        internal_print_generic(
            *args, raise_on_nan=raise_on_nan, as_leaves=as_leaves, console=console
        )
        return

    leaves, tree_def = tree.flatten(args)

    def fix(u_arg: object, arg: object) -> object:
        if isinstance(arg, int):
            assert isinstance(u_arg, SupportsInt)
            return int(u_arg)
        return np.asarray(u_arg) if isinstance(arg, np.ndarray) else u_arg

    def callback(*callback_leaves: object) -> None:
        unflattened_tree = tree.unflatten(tree_def, callback_leaves)
        v_args = tree.map(fix, unflattened_tree, args)
        internal_print_generic(
            *v_args, raise_on_nan=raise_on_nan, as_leaves=as_leaves, console=console
        )

    debug.callback(callback, *leaves, ordered=True)



from collections.abc import Generator
from io import StringIO

import numpy as np
import pytest
from jax import enable_x64
from rich.console import Console


@pytest.fixture(autouse=True)
def _jax_enable64() -> Generator[None]:  # pyright: ignore
    with enable_x64():
        yield


@pytest.fixture
def generator() -> np.random.Generator:
    return np.random.default_rng(123)


@pytest.fixture
def console() -> Console:
    return Console(file=StringIO(), width=80)



"""Avoiding recompilation.

Using tjax.gradient is much more efficient than optax since we can pass gradient transformation
objects without inducing unnecessary recompilation.
"""

from functools import partial
from typing import Any

from jax import jit
from optax import GradientTransformation as OptaxGradientTransformation
from optax import adam

from tjax.gradient import Adam, GradientTransformation


@partial(jit, static_argnames=("optax_gt",))
def try_model_optax(optax_gt: OptaxGradientTransformation) -> None:
    print("Recompiling using optax!")  # noqa: T201


@jit
def try_model_tjax(tjax_gt: GradientTransformation[Any, Any]) -> None:
    print("Recompiling using tjax!")  # noqa: T201


for learning_rate in [0.1, 0.2, 0.4, 0.8]:
    # Typically, the model and gradient transformations are created here.
    optax_gt = adam(learning_rate)
    tjax_gt = Adam(learning_rate)
    # Now, you would create the gradient state.

    # Then, these are passed to a jitted inference function.
    try_model_optax(optax_gt)  # type: ignore
    try_model_tjax(tjax_gt)

    # With the result of infrence, you would update the grdient state.

# Prints:
# Recompiling using optax!
# Recompiling using tjax!
# Recompiling using optax!
# Recompiling using optax!
# Recompiling using optax!

# To avoid recompilation in Optax, you are forced to create the gradient transformation again inside
# the inference function.



from __future__ import annotations

from collections.abc import Callable
from typing import Any, override

from .function_markers import abstract_custom_jvp_marker, abstract_jit_marker
from .shims import custom_jvp_method, jit


class JaxAbstractClass:
    """A class with abstract methods whose overrides need to be decorated with Jax decorators."""

    @override
    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        super_cls = super(cls, cls)
        for name, original_method in list(cls.__dict__.items()):
            if not hasattr(super_cls, name):
                continue
            super_method = getattr(super_cls, name)
            if super_method is original_method:
                continue
            # Override detected.
            method = original_method
            if (kwargs := getattr(super_method, abstract_jit_marker, None)) is not None:
                method = jit(original_method, **kwargs)
            if hasattr(super_method, abstract_custom_jvp_marker):
                jvp, nondiff_argnums = super_method._abstract_custom_jvp  # noqa: SLF001
                method_jvp: Any = custom_jvp_method(method, nondiff_argnums=nondiff_argnums)
                method_jvp.defjvp(jvp)
                setattr(cls, f"_original_{name}", method)
                method = method_jvp
            setattr(cls, name, method)


def abstract_jit[F: Callable[..., Any]](fun: F, **kwargs: object) -> F:
    """An abstract method whose override need to be jitted."""
    setattr(fun, abstract_jit_marker, kwargs)
    return fun


def abstract_custom_jvp[**P, R_co](
    jvp: Callable[..., tuple[R_co, R_co]], nondiff_argnums: tuple[int, ...] = ()
) -> Callable[[Callable[P, R_co]], Callable[P, R_co]]:
    """An abstract method whose override need to be decorated with custom_jvp_method."""

    def decorator(fun: Callable[P, R_co]) -> Callable[P, R_co]:
        fun._abstract_custom_jvp = (  # type: ignore # pyright: ignore  # noqa: SLF001
            jvp,
            nondiff_argnums,
        )
        return fun

    return decorator  # type: ignore



from ._src.dataclasses.dataclass import DataclassInstance, TDataclassInstance, dataclass
from ._src.dataclasses.helpers import as_shallow_dict, field

__all__ = ["DataclassInstance", "TDataclassInstance", "as_shallow_dict", "dataclass", "field"]



from __future__ import annotations

from collections.abc import Callable
from typing import override

import jax.numpy as jnp
from jax import tree

from tjax.dataclasses import dataclass

from tjax._src.annotations import ComplexNumeric, JaxComplexArray, PyTree, RealNumeric
from .transform import GradientState, SecondOrderGradientTransformation


@dataclass
class SMDState[Weights: PyTree](GradientState):
    log_learning_rate: Weights
    v: Weights


@dataclass
class SMDGradient[Weights: PyTree](SecondOrderGradientTransformation[SMDState[Weights], Weights]):
    """Stochastic meta-descent.

    Schraudolph, N. N. (1999). Local gain adaptation in stochastic gradient descent. Artificial
    Neural Networks, 1999. ICANN 99. Ninth International Conference on (Conf. Publ. No. 470), 2,
    569-574. https://doi.org/10.1049/cp:19991170.
    """

    meta_learning_rate: RealNumeric = 1e-2

    @override
    def init(self, parameters: Weights) -> SMDState[Weights]:
        z = tree.map(jnp.zeros_like, parameters)
        return SMDState[Weights](z, z)

    @override
    def second_order_update(
        self,
        gradient: Weights,
        state: SMDState[Weights],
        parameters: Weights | None,
        hessian_vector_product: Callable[[Weights], Weights],
    ) -> tuple[Weights, SMDState[Weights]]:
        negative_gradient = tree.map(jnp.negative, gradient)  # delta

        # Update log-learning rate.
        def g(log_p: RealNumeric, delta: ComplexNumeric, v: ComplexNumeric) -> JaxComplexArray:
            return jnp.asarray(log_p + self.meta_learning_rate * delta * v)

        new_log_learning_rate = tree.map(g, state.log_learning_rate, negative_gradient, state.v)
        learning_rate = tree.map(jnp.exp, new_log_learning_rate)  # p

        # Calculate gradient.
        gradient = tree.map(jnp.multiply, learning_rate, negative_gradient)

        # Update v.
        def f(
            v: ComplexNumeric, p: RealNumeric, delta: ComplexNumeric, hv: ComplexNumeric
        ) -> JaxComplexArray:
            return jnp.asarray(v + p * delta - hv)

        new_v = tree.map(
            f, state.v, learning_rate, negative_gradient, hessian_vector_product(state.v)
        )

        return gradient, SMDState[Weights](new_log_learning_rate, new_v)


# got
# Union[ndarray[Any, dtype[complexfloating[Any, Any]]],
#       ndarray[Any, dtype[floating[Any]]],
#       ndarray[Any, dtype[signedinteger[Any]]],
#       Array,
#       complex]
# , expected
# Union[ndarray[Any, dtype[Union[floating[Any], complexfloating[Any, Any]]]],
#       Array,
#       complex, float, int]



from __future__ import annotations

import dataclasses
from collections.abc import Callable, Mapping
from dataclasses import MISSING, fields
from typing import Any, ClassVar, Never, Protocol, overload, runtime_checkable


@runtime_checkable
class DataclassInstance(Protocol):
    __dataclass_fields__: ClassVar[dict[str, dataclasses.Field[Any]]]


# NOTE: Actual return type is 'Field[T]', but we want to help type checkers
# to understand the magic that happens at runtime.
@overload  # `default` and `default_factory` are optional and mutually exclusive.
def field[T](
    *,
    static: bool = False,
    default: T,
    init: bool = ...,
    repr: bool = ...,
    hash: bool | None = ...,
    compare: bool = ...,
    metadata: Mapping[str, Any] | None = ...,
    kw_only: bool = ...,
) -> T: ...


@overload
def field[T](
    *,
    static: bool = False,
    default_factory: Callable[[], T],
    init: bool = ...,
    repr: bool = ...,
    hash: bool | None = ...,
    compare: bool = ...,
    metadata: Mapping[str, Any] | None = ...,
    kw_only: bool = ...,
) -> T: ...


@overload
def field(
    *,
    static: bool = False,
    init: bool = ...,
    repr: bool = ...,
    hash: bool | None = ...,
    compare: bool = ...,
    metadata: Mapping[str, Any] | None = ...,
    kw_only: bool = ...,
) -> Never: ...


def field(
    *,
    static: bool = False,
    default: Any = MISSING,
    default_factory: Any = MISSING,
    init: bool = True,
    repr: bool = True,  # noqa: A002
    hash: bool | None = None,  # noqa: A002
    compare: bool = True,
    metadata: Mapping[str, Any] | None = None,
    kw_only: Any = MISSING,
) -> Any:
    """A field creator with a static indicator.

    The static flag indicates whether a field is a pytree or static.  Pytree fields are
    differentiated and traced.  Static fields are hashed and compared.
    """
    metadata_dict: dict[str, Any] = {} if metadata is None else dict(metadata)
    metadata_dict["static"] = static
    if default is MISSING:
        return dataclasses.field(
            default_factory=default_factory,
            init=init,
            repr=repr,
            hash=hash,
            compare=compare,
            metadata=metadata_dict,
            kw_only=kw_only,
        )
    return dataclasses.field(
        default=default,
        init=init,
        repr=repr,
        hash=hash,
        compare=compare,
        metadata=metadata_dict,
        kw_only=kw_only,
    )


def as_shallow_dict(dcls: DataclassInstance) -> dict[str, Any]:
    return {field.name: getattr(dcls, field.name) for field in fields(dcls)}



from __future__ import annotations

import logging
from contextlib import AbstractContextManager
from time import perf_counter_ns
from types import TracebackType
from typing import Self, override

_metric_factor = 1000
log = logging.getLogger(__name__)


class Timer(AbstractContextManager["Timer"]):
    @override
    def __init__(self, final_string: str, precision: int = 3) -> None:
        super().__init__()
        self.start = 0
        self.end = 0
        self.final_string = final_string
        self.precision = precision

    @override
    def __enter__(self) -> Self:
        _ = super().__enter__()
        self.start = perf_counter_ns()
        return self

    @override
    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
        /,
    ) -> None:
        super().__exit__(exc_type, exc_val, exc_tb)  # pyright: ignore
        self.end = perf_counter_ns()
        log.info(f"{self.final_string} took {self.elapsed_str()}")

    def elapsed_ns(self) -> int:
        return self.end - self.start

    def elapsed_str(self) -> str:
        return display_time(self.elapsed_ns(), precision=self.precision)


def display_time(elapsed_ns: float, *, precision: int = 3) -> str:
    elapsed = elapsed_ns
    for unit in ["n", "μ", "m"]:  # noqa: B007
        if elapsed < _metric_factor:
            break
        elapsed /= _metric_factor
    else:
        unit = ""
    if elapsed >= 1:
        precision = max(0, precision - len(str(int(elapsed))))
    return f"{elapsed:.{precision}f} {unit}s"



from __future__ import annotations

import dataclasses
from collections.abc import Callable
from typing import Any, ClassVar, Protocol, cast, dataclass_transform, overload, runtime_checkable

from jax.tree_util import register_dataclass

from tjax._src.testing import get_relative_test_string, get_test_string, tree_allclose
from .helpers import DataclassInstance, field


@runtime_checkable
class TDataclassInstance(DataclassInstance, Protocol):
    static_fields: ClassVar[list[str]]
    dynamic_fields: ClassVar[list[str]]


@overload
@dataclass_transform(frozen_default=True, field_specifiers=(field,))
def dataclass(
    *,
    init: bool = True,
    repr: bool = True,
    eq: bool = True,
    order: bool = False,
    frozen: bool = True,
) -> Callable[[type[Any]], type[TDataclassInstance]]: ...
@overload
@dataclass_transform(frozen_default=True, field_specifiers=(field,))
def dataclass(
    cls: type[Any],
    /,
    *,
    init: bool = True,
    repr: bool = True,
    eq: bool = True,
    order: bool = False,
    frozen: bool = True,
) -> type[TDataclassInstance]: ...
@dataclass_transform(frozen_default=True, field_specifiers=(field,))
def dataclass(
    cls: type[Any] | None = None,
    /,
    *,
    init: bool = True,
    repr: bool = True,  # noqa: A002
    eq: bool = True,
    order: bool = False,
    frozen: bool = True,
) -> type[TDataclassInstance] | Callable[[type[Any]], type[TDataclassInstance]]:
    """A dataclass creator that creates a Jax pytree.

    Returns the same class as was passed in, with dunder methods added based on the fields defined
    in the class.

    Examines PEP 526 annotations to determine fields.  Default values for fields are provided using
    assignment.

    To mark fields as static fields rather than JAX pytree fields, use the `field` function.
    In JAX, a static attribute is one that induces recompilation of a function when it changes, and
    consequently there is more flexibility about what can be done with such an attribute.

    For example::
    ```python
    from __future__ import annotations

    from typing import ClassVar

    import jax.numpy as jnp
    from tjax.dataclasses import dataclass, field
    from jax import Array, grad

    @dataclass
    class LearnedParameter:
        weight: Array
        constrain_positive: bool = field(static=True)
        minimum_positive_weight: ClassVar[Array] = 1e-6

        def trained(self,
                    self_bar: LearnedParameter,
                    learning_rate: float) -> LearnedParameter:
            weight_bar = self_bar.weight
            weight = self.weight - weight_bar * learning_rate
            if self.constrain_positive:
                weight = jnp.maximum(weight, self.minimum_positive_weight)
            return LearnedParameter(weight=weight,
                                    constrain_positive=self.constrain_positive)

    def loss(w: LearnedParameter) -> float:
        return jnp.square(w.weight - 3.3)

    w = LearnedParameter(2.0, True)
    w_bar = grad(loss)(w)
    new_w = w.trained(w_bar, 1e-4)
    ```

    Since this dataclass is a pytree, all of JAX's functions that accept pytrees work with it,
    including iteration, differentiation, and `jax.tree_util` functions.
    """
    if cls is None:

        def f(x: type[Any], /) -> type[TDataclassInstance]:
            return dataclass(x, init=init, repr=repr, eq=eq, order=order, frozen=frozen)  # type: ignore

        return f  # Type checking support partial is poor.

    # Apply dataclass function to cls.
    data_clz: type[TDataclassInstance] = cast(
        "type[TDataclassInstance]",
        dataclasses.dataclass(init=init, repr=repr, eq=eq, order=order, frozen=frozen)(cls),
    )

    # Partition fields into static, and dynamic; and assign these to the class.
    static_fields: list[str] = []
    dynamic_fields: list[str] = []
    for field_info in dataclasses.fields(data_clz):
        if not field_info.init:
            continue
        if field_info.metadata.get("static", False):
            static_fields.append(field_info.name)
        else:
            dynamic_fields.append(field_info.name)
    data_clz.static_fields = static_fields
    data_clz.dynamic_fields = dynamic_fields
    _ = register_dataclass(data_clz, dynamic_fields, static_fields)

    # Register the dynamically-dispatched functions.
    _ = get_test_string.register(data_clz, get_dataclass_test_string)
    _ = get_relative_test_string.register(data_clz, get_relative_dataclass_test_string)
    return data_clz


def get_dataclass_test_string(actual: DataclassInstance, rtol: float, atol: float) -> str:
    retval = f"{type(actual).__name__}("
    retval += ",\n".join(
        (
            (f"{field_object.name}=" if field_object.kw_only else "")
            + get_test_string(getattr(actual, field_object.name), rtol, atol)
        )
        for field_object in dataclasses.fields(actual)
        if field_object.repr
    )
    retval += ")"
    return retval


def get_relative_dataclass_test_string(
    actual: TDataclassInstance, original_name: str, original: object, rtol: float, atol: float
) -> str:
    repr_dict = {field.name: field.repr for field in dataclasses.fields(actual)}
    retval = f"replace({original_name}, "
    retval += ",\n".join(
        f"{field_name}="
        + get_relative_test_string(
            getattr(actual, field_name),
            f"{original_name}.{field_name}",
            getattr(original, field_name),
            rtol,
            atol,
        )
        for field_name in actual.dynamic_fields
        if repr_dict[field_name]
        if not tree_allclose(
            getattr(actual, field_name), getattr(original, field_name), rtol=rtol, atol=atol
        )
    )
    retval += ")"
    return retval



from __future__ import annotations

from functools import partial

import jax.numpy as jnp
from jax import Array, vjp
from numpy.testing import assert_equal

from tjax import (
    JaxRealArray,
    assert_tree_allclose,
    copy_cotangent,
    cotangent_combinator,
    replace_cotangent,
    reverse_scale_cotangent,
    scale_cotangent,
)
from tjax.dataclasses import dataclass


@dataclass
class X:
    x: JaxRealArray
    y: JaxRealArray


def test_scalar_scale() -> None:
    o = jnp.ones(())
    x = X(2 * o, 3 * o)
    x_out, vjp_f = vjp(partial(scale_cotangent, scalar_scale=3.0), x)
    assert_tree_allclose(x, x_out)
    (x_bar,) = vjp_f(X(4 * o, 5 * o))
    assert_tree_allclose(x_bar, X(12 * o, 15 * o))


def test_tree_scale() -> None:
    o = jnp.ones(())
    x = X(2 * o, 3 * o)
    tree_scale = X(6 * o, -7 * o)
    x_out, vjp_f = vjp(partial(scale_cotangent, tree_scale=tree_scale), x)
    assert_tree_allclose(x, x_out)
    (x_bar,) = vjp_f(X(4 * o, 5 * o))
    assert_tree_allclose(x_bar, X(24 * o, -35 * o))


def test_scalar_tree_scale() -> None:
    o = jnp.ones(())
    x = X(2 * o, 3 * o)
    tree_scale = X(6 * o, -7 * o)
    x_out, vjp_f = vjp(partial(scale_cotangent, scalar_scale=3.0, tree_scale=tree_scale), x)
    assert_tree_allclose(x, x_out)
    (x_bar,) = vjp_f(X(4 * o, 5 * o))
    assert_tree_allclose(x_bar, X(3 * 24 * o, 3 * -35 * o))


def test_reverse_scale() -> None:
    z = jnp.zeros(())
    o = jnp.ones(())
    x = X(2 * o, 3 * o)
    (x_out, zero), vjp_f = vjp(reverse_scale_cotangent, x)
    assert_tree_allclose(x, x_out)
    assert_tree_allclose(z, zero)
    (x_bar,) = vjp_f((X(4 * o, 5 * o), 2.0))
    assert_tree_allclose(x_bar, X(8 * o, 10 * o))


def test_replace_cotangent() -> None:
    primals, vjp_f = vjp(replace_cotangent, 1.0, 2.0)
    assert_equal(primals, 1.0)
    assert_equal(vjp_f(3.0), (2.0, 3.0))


def test_copy_cotangent() -> None:
    primals, vjp_f = vjp(copy_cotangent, 1.0, 2.0)
    assert_equal(primals, 1.0)
    assert_equal(vjp_f(3.0), (3.0, 3.0))


def test_combinator() -> None:
    def f(x: float | Array) -> tuple[tuple[float | Array, float | Array], None]:
        return (x**2, x**2), None

    o = jnp.ones(())

    args = ((-1.0,), (-1.0,))
    _, f_vjp = vjp(partial(cotangent_combinator, f, aux_cotangent_scales=None), args)
    result_bar = (2 * o, 3 * o), None
    (actual_args_bar,) = f_vjp(result_bar)
    desired_args_bar = ((-4.0,), (-6.0,))
    assert_tree_allclose(actual_args_bar, desired_args_bar)



from __future__ import annotations

from collections.abc import Callable
from typing import Any

import chex
import jax.numpy as jnp
import pytest
from jax import tree

from tjax.gradient import (
    DPSGD,
    LARS,
    LBFGS,
    SGD,
    SM3,
    AdaBelief,
    AdaDelta,
    AdaFactor,
    AdaGrad,
    Adam,
    Adamax,
    AdamaxW,
    AdamW,
    Fromage,
    Lamb,
    Lion,
    NoisySGD,
    Novograd,
    OptimisticGradientDescent,
    RAdam,
    RMSProp,
    Schedule,
    Yogi,
)

Params = dict[str, jnp.ndarray]


def _params() -> Params:
    return {"w": jnp.array([1.0, 2.0])}


def _gradient() -> Params:
    return {"w": jnp.array([0.1, 0.2])}


@pytest.mark.parametrize(
    "constructor",
    [
        lambda: AdaBelief[Params](1e-2),
        AdaDelta[Params],
        AdaFactor[Params],
        lambda: AdaGrad[Params](1e-2),
        lambda: Adam[Params](1e-2),
        lambda: Adam[Params](Schedule(lambda _: 1e-2)),
        lambda: AdamW[Params](1e-2),
        lambda: Adamax[Params](1e-2),
        lambda: AdamaxW[Params](1e-2),
        lambda: Fromage[Params](1e-2),
        lambda: Lamb[Params](1e-2),
        lambda: LARS[Params](1e-2),
        lambda: Lion[Params](1e-2),
        lambda: NoisySGD[Params](1e-2),
        lambda: Novograd[Params](1e-2),
        lambda: OptimisticGradientDescent[Params](1e-2, 1e-2, 1e-2),
        lambda: RAdam[Params](1e-2),
        lambda: RMSProp[Params](1e-2),
        lambda: SGD[Params](1e-2),
        lambda: SM3[Params](1e-2),
        lambda: Yogi[Params](1e-2),
        lambda: DPSGD[Params](1e-2, 1.0, 0.0, 0),
    ],
)
def test_aliases_smoke(
    constructor: Callable[[], Any],
) -> None:
    params = _params()
    gradient = _gradient()
    tx = constructor()
    state = tx.init(params)
    updates, new_state = tx.update(gradient, state, params)

    chex.assert_tree_all_finite(updates)
    assert type(new_state) is type(state)
    assert tree.structure(updates) == tree.structure(gradient)


def test_lbfgs_init() -> None:
    state = LBFGS[Params]().init(_params())
    assert state is not None



from __future__ import annotations

from collections.abc import MutableSet
from typing import Any

from rich.tree import Tree

from tjax._src.display.display_generic import _verify, display_class, display_generic


def graph_arrow(directed: bool) -> str:  # noqa: FBT001
    return "⟶" if directed else "↔"


def graph_edge_name(arrow: str, source: str, target: str) -> str:
    return f"{source}{arrow}{target}"


try:
    import networkx as nx
except ImportError:
    pass
else:

    @display_generic.register(nx.Graph)
    def _(
        value: nx.Graph[Any],
        *,
        seen: MutableSet[int] | None = None,
        key: str = "",
    ) -> Tree:
        if seen is None:
            seen = set()
        with _verify(value, seen, key) as x:
            if x:
                return x
            arrow = graph_arrow(isinstance(value, nx.DiGraph))
            retval = display_class(key, type(value))
            for name, node in value.nodes.items():
                retval.children.append(display_generic(node, seen=seen, key=name))
            for (source, target), edge in value.edges.items():
                key = graph_edge_name(arrow, source, target)
                retval.children.append(display_generic(edge, seen=seen, key=key))
            return retval



from __future__ import annotations

from typing import overload

import jax.numpy as jnp
import jax.random as jr
from array_api_compat import array_namespace

from .annotations import (
    ComplexArray,
    IntegralArray,
    JaxComplexArray,
    JaxIntegralArray,
    JaxRealArray,
    KeyArray,
    RealArray,
)


@overload
def leaky_integrate(
    value: JaxComplexArray,
    time_step: JaxRealArray,
    drift: JaxComplexArray | None = None,
    decay: JaxComplexArray | None = None,
    *,
    leaky_average: bool = False,
) -> JaxComplexArray: ...


@overload
def leaky_integrate(
    value: ComplexArray,
    time_step: RealArray,
    drift: ComplexArray | None = None,
    decay: ComplexArray | None = None,
    *,
    leaky_average: bool = False,
) -> ComplexArray: ...


def leaky_integrate(
    value: ComplexArray,
    time_step: RealArray,
    drift: ComplexArray | None = None,
    decay: ComplexArray | None = None,
    *,
    leaky_average: bool = False,
) -> ComplexArray:
    """Update the value so that it is the leaky integral (or leaky average).

    Args:
        value: The current value of the leaky integral or average.
        time_step: The number of seconds that have passed.
        decay: If provided, must have positive real component, and the value decays by exp(-decay)
            every second.
        drift: If provided, the value increases by this every second.
        leaky_average: A flag indicating a leaky average rather than a leaky integral.  This scales
            the drift by the real component (in case the decay is complex) of the decay.
    """
    if drift is None:
        if decay is None:
            return value
        xp = array_namespace(decay, time_step, value)
        return xp.exp(-decay * time_step) * value

    if decay is None:
        if leaky_average:
            raise ValueError
        xp = array_namespace(drift, time_step, value)
        return xp.asarray(value + drift * time_step)

    xp = array_namespace(drift, decay, time_step, value)
    scaled_integrand = (drift / decay) * -xp.expm1(-decay * time_step)

    if leaky_average:
        scaled_integrand *= decay.real

    return xp.exp(-decay * time_step) * value + scaled_integrand


@overload
def leaky_data_weight(
    iterations_times_time_step: JaxRealArray | JaxIntegralArray, decay: JaxRealArray
) -> JaxRealArray: ...


@overload
def leaky_data_weight(
    iterations_times_time_step: RealArray | IntegralArray, decay: RealArray
) -> RealArray: ...


def leaky_data_weight(
    iterations_times_time_step: RealArray | IntegralArray, decay: RealArray
) -> RealArray:
    """The amount of data that has been incorporated and has not been decayed.

    This equals leaky_integrate(0.0, iterations_times_time_step, 1.0, decay, leaky_average=True).
    """
    xp = array_namespace(iterations_times_time_step, decay)
    return -xp.expm1(-iterations_times_time_step * decay)


def diffused_leaky_integrate(
    value: JaxComplexArray,
    time_step: JaxRealArray,
    rng: KeyArray,
    diffusion: JaxRealArray,
    drift: JaxComplexArray | None = None,
    decay: JaxComplexArray | None = None,
    *,
    leaky_average: bool = False,
) -> ComplexArray:
    """Update an Ornstein-Uhlenbeck process.

    Args:
        value: The current value of the leaky integral or average.
        time_step: The number of seconds that have passed.
        rng: The key array for the stochastic process.
        diffusion: The diffusion for the stochastic process.
        decay: If provided, the value decays by exp(-decay) every second.
        drift: If provided, the value increases by this every second.
        leaky_average: A flag indicating a leaky average rather than a leaky integral.  This scales
            the drift by the real component (in case the decay is complex) of the decay.
    """
    variance = (
        diffusion * time_step
        if decay is None
        else diffusion / decay * -jnp.expm1(-decay * time_step)
    )
    jump = jnp.sqrt(variance) * jr.normal(rng, value.shape)
    return leaky_integrate(value, time_step, drift, decay, leaky_average=leaky_average) + jump



"""Meta-learning.

Using tjax.gradient is much simpler than optax since we don't need inject_hyperparams.

Compare this example with
https://optax.readthedocs.io/en/latest/_collections/examples/meta_learning.html
"""

import operator
from typing import Any

import jax
import jax.numpy as jnp
import jax.random as jr
import jax.scipy.special as jss
import matplotlib.pyplot as plt
from jax import tree

from tjax import JaxArray, RngStream
from tjax.dataclasses import dataclass
from tjax.gradient import Adam, GenericGradientState, GradientState, GradientTransformation, RMSProp


@dataclass
class InferenceState:
    theta: JaxArray  # Model parameters.
    state: GradientState  # Gradient state of the model parameters.


@dataclass
class OuterInferenceState:
    inference_state: InferenceState
    eta: JaxArray  # Hyperparameters.
    meta_state: GenericGradientState  # Gradient state of the hyperparameters.


@dataclass
class TrainingExample:
    observation: JaxArray
    target: JaxArray


def apply_updates[T](parameters: T, updates: T) -> T:
    return tree.map(operator.add, parameters, updates)


def generate_example(stream: RngStream) -> TrainingExample:
    observation = jr.uniform(stream.key(), minval=0.0, maxval=10.0)
    target = 10.0 * observation + jr.normal(stream.key())
    return TrainingExample(observation, target)


def model(theta: JaxArray, x: JaxArray) -> JaxArray:
    # This is the simplest possible model.
    return x * theta


def loss(theta: JaxArray, training_example: TrainingExample) -> JaxArray:
    return jnp.sum(jnp.square(training_example.target - model(theta, training_example.observation)))


def step(
    opt: GradientTransformation[Any, JaxArray],
    inference_state: InferenceState,
    training_example: TrainingExample,
) -> InferenceState:
    gradient = jax.grad(loss)(inference_state.theta, training_example)
    updates, state = opt.update(gradient, inference_state.state, inference_state.theta)
    theta = apply_updates(inference_state.theta, updates)
    return InferenceState(theta, state)


@jax.jit
def outer_loss(
    eta: JaxArray,
    inference_state: InferenceState,
    training_examples: list[TrainingExample],
) -> tuple[JaxArray, InferenceState]:
    # Eta is the logit of the learning rate.
    opt = RMSProp(learning_rate=jss.expit(eta))
    # Use all but one sample to train theta.
    for training_example in training_examples[:-1]:
        inference_state = step(opt, inference_state, training_example)
    # Use the final sample to train eta.
    training_example = training_examples[-1]
    return loss(inference_state.theta, training_example), inference_state


@jax.jit
def outer_step(
    outer_inference_state: OuterInferenceState,
    training_examples: list[tuple[JaxArray, JaxArray]],
    meta_opt: GradientTransformation[Any, JaxArray],
) -> OuterInferenceState:
    g_outer_loss = jax.grad(outer_loss, has_aux=True)
    gradient, inference_state = g_outer_loss(
        outer_inference_state.eta, outer_inference_state.inference_state, training_examples
    )

    meta_updates, meta_state = meta_opt.update(gradient, outer_inference_state.meta_state, None)
    new_eta = apply_updates(outer_inference_state.eta, meta_updates)
    return OuterInferenceState(inference_state, new_eta, meta_state)


def create_state() -> tuple[OuterInferenceState, GradientTransformation[Any, JaxArray]]:
    init_learning_rate = jnp.array(0.1)
    meta_learning_rate = jnp.array(0.03)
    opt = RMSProp(learning_rate=init_learning_rate)
    meta_opt = Adam(learning_rate=meta_learning_rate)
    # Eta is the logit of the learning rate.
    eta = jss.logit(init_learning_rate)
    theta = jax.random.normal(jax.random.PRNGKey(42))
    inference_state = InferenceState(theta=theta, state=opt.init(theta))
    outer_inference_state = OuterInferenceState(
        inference_state=inference_state, eta=eta, meta_state=meta_opt.init(eta)
    )
    return outer_inference_state, meta_opt


def run() -> None:
    outer_inference_state, meta_opt = create_state()
    stream = RngStream(jax.random.key(0))
    learning_rates = []
    thetas = []

    for _ in range(2000):
        training_examples = [generate_example(stream) for _ in range(7)]
        outer_inference_state = outer_step(outer_inference_state, training_examples, meta_opt)
        learning_rates.append(jss.expit(outer_inference_state.eta))
        thetas.append(outer_inference_state.inference_state.theta)

    fig, (ax1, ax2) = plt.subplots(2)
    fig.suptitle("Meta-learning RMSProp's learning rate")
    plt.xlabel("Step")

    ax1.semilogy(range(len(learning_rates)), learning_rates)
    ax1.set(ylabel="Learning rate")
    ax1.label_outer()

    plt.xlabel("Number of updates")
    ax2.semilogy(range(len(thetas)), thetas)

    ax2.label_outer()
    ax2.set(ylabel="Theta")

    plt.show()


run()



"""Tools for the differential programming library [JAX](https://github.com/google/jax)."""

from . import dataclasses, gradient
from ._src.abstract_method_decorators import JaxAbstractClass, abstract_custom_jvp, abstract_jit
from ._src.annotations import (
    Array,
    BooleanArray,
    BooleanNumeric,
    ComplexArray,
    ComplexNumeric,
    IntegralArray,
    IntegralNumeric,
    JaxArray,
    JaxBooleanArray,
    JaxComplexArray,
    JaxIntegralArray,
    JaxRealArray,
    KeyArray,
    NumpyArray,
    NumpyBooleanArray,
    NumpyBooleanNumeric,
    NumpyComplexArray,
    NumpyComplexNumeric,
    NumpyIntegralArray,
    NumpyIntegralNumeric,
    NumpyRealArray,
    NumpyRealNumeric,
    PyTree,
    RealArray,
    RealNumeric,
    Shape,
    ShapeLike,
    SliceLike,
)
from ._src.cotangent_tools import (
    copy_cotangent,
    cotangent_combinator,
    print_cotangent,
    replace_cotangent,
    reverse_scale_cotangent,
    scale_cotangent,
    zero_from_primal,
)
from ._src.display.display_generic import display_generic
from ._src.display.generic_string import GenericString
from ._src.display.internal import internal_print_generic
from ._src.display.print_generic import print_generic
from ._src.dtype_tools import cast_to_result_type
from ._src.graph.display import graph_arrow, graph_edge_name
from ._src.graph.register_jax import register_graph_as_jax_pytree
from ._src.graph.types_ import GraphEdgeKey, GraphNodeKey, UndirectedGraphEdgeKey
from ._src.leaky_integral import diffused_leaky_integrate, leaky_data_weight, leaky_integrate
from ._src.math_tools import (
    abs_square,
    bessel_iv_ratio,
    divide_where,
    inverse_softplus,
    log_bessel_ive,
    log_softplus,
    matrix_dot_product,
    matrix_vector_mul,
    normalize,
    outer_product,
    softplus,
    stop_gradient,
    sublinear_softplus,
)
from ._src.project_tree import Projectable, Projector
from ._src.rng import RngStream, create_streams, fork_streams, sample_streams
from ._src.shims import custom_jvp, custom_jvp_method, custom_vjp, custom_vjp_method, hessian, jit
from ._src.testing import (
    assert_tree_allclose,
    get_relative_test_string,
    get_test_string,
    tree_allclose,
)
from ._src.timer import Timer, display_time
from ._src.tree_tools import dynamic_tree_all, element_count, tree_sum

__all__ = [
    "Array",
    "BooleanArray",
    "BooleanNumeric",
    "ComplexArray",
    "ComplexNumeric",
    "GenericString",
    "GraphEdgeKey",
    "GraphNodeKey",
    "IntegralArray",
    "IntegralNumeric",
    "JaxAbstractClass",
    "JaxArray",
    "JaxBooleanArray",
    "JaxComplexArray",
    "JaxIntegralArray",
    "JaxRealArray",
    "KeyArray",
    "NumpyArray",
    "NumpyBooleanArray",
    "NumpyBooleanNumeric",
    "NumpyComplexArray",
    "NumpyComplexNumeric",
    "NumpyIntegralArray",
    "NumpyIntegralNumeric",
    "NumpyRealArray",
    "NumpyRealNumeric",
    "Projectable",
    "Projector",
    "PyTree",
    "RealArray",
    "RealNumeric",
    "RngStream",
    "Shape",
    "ShapeLike",
    "SliceLike",
    "Timer",
    "UndirectedGraphEdgeKey",
    "abs_square",
    "abstract_custom_jvp",
    "abstract_jit",
    "assert_tree_allclose",
    "bessel_iv_ratio",
    "cast_to_result_type",
    "copy_cotangent",
    "cotangent_combinator",
    "create_streams",
    "custom_jvp",
    "custom_jvp_method",
    "custom_vjp",
    "custom_vjp_method",
    "dataclasses",
    "diffused_leaky_integrate",
    "display_generic",
    "display_time",
    "divide_where",
    "dynamic_tree_all",
    "element_count",
    "fork_streams",
    "get_relative_test_string",
    "get_test_string",
    "gradient",
    "graph_arrow",
    "graph_edge_name",
    "hessian",
    "internal_print_generic",
    "inverse_softplus",
    "jit",
    "leaky_data_weight",
    "leaky_integrate",
    "log_bessel_ive",
    "log_softplus",
    "matrix_dot_product",
    "matrix_vector_mul",
    "normalize",
    "outer_product",
    "print_cotangent",
    "print_generic",
    "register_graph_as_jax_pytree",
    "replace_cotangent",
    "reverse_scale_cotangent",
    "sample_streams",
    "scale_cotangent",
    "softplus",
    "stop_gradient",
    "sublinear_softplus",
    "tree_allclose",
    "tree_sum",
    "zero_from_primal",
]



"""Tests from optax._src.transform_test."""

from collections.abc import Callable
from typing import Any

import chex
import jax.numpy as jnp
import numpy as np
import pytest
from jax import Array, tree
from optax import apply_updates

from tjax import RealArray
from tjax.gradient import (
    SGD,
    AddDecayedWeights,
    ApplyEvery,
    Centralize,
    ChainedGradientTransformation,
    Ema,
    GradientTransformation,
    Scale,
    ScaleByAdam,
    ScaleByParamBlockNorm,
    ScaleByParamBlockRMS,
    ScaleByRms,
    ScaleByStddev,
    ScaleByTrustRatio,
    Trace,
)

STEPS = 50
LR = 1e-2
init_params = (np.asarray([1.0, 2.0]), np.asarray([3.0, 4.0]))
per_step_updates = (np.asarray([500.0, 5.0]), np.asarray([300.0, 3.0]))


def variant[T](x: T) -> T:
    return x


@pytest.mark.parametrize(
    "constructor",
    [
        ScaleByAdam,
        ScaleByRms,
        ScaleByStddev,
        ScaleByTrustRatio,
        ScaleByParamBlockNorm,
        ScaleByParamBlockRMS,
    ],
)
def test_scalers(constructor: Callable[[], GradientTransformation[Any, Any]]) -> None:
    params = init_params

    scaler = constructor()
    init_fn = variant(scaler.init)
    transform_fn = variant(scaler.update)

    state = init_fn(params)
    chex.assert_tree_all_finite(state)

    updates, state = transform_fn(per_step_updates, state, params)
    chex.assert_tree_all_finite((params, updates, state))
    tree.map(lambda *args: chex.assert_equal_shape(args), params, updates)


def test_add_decayed_weights() -> None:
    # Define a transform that add decayed weights.
    # We can define a mask either as a pytree, or as a function that
    # returns the pytree. Below we define the pytree directly.
    mask = (True, {"a": True, "b": False})
    tx = AddDecayedWeights[Any](0.1, mask=mask)
    # Define input updates and weights.
    updates = (
        jnp.zeros((2,), dtype=jnp.float32),
        {"a": jnp.zeros((2,), dtype=jnp.float32), "b": jnp.zeros((2,), dtype=jnp.float32)},
    )
    weights = (
        jnp.ones((2,), dtype=jnp.float32),
        {"a": jnp.ones((2,), dtype=jnp.float32), "b": jnp.ones((2,), dtype=jnp.float32)},
    )
    # This mask means that we will add decayed weights to the first two
    # terms in the input updates, but not to the last element.
    expected_tx_updates = (
        0.1 * jnp.ones((2,), dtype=jnp.float32),
        {"a": 0.1 * jnp.ones((2,), dtype=jnp.float32), "b": jnp.zeros((2,), dtype=jnp.float32)},
    )
    # Apply transform
    state = tx.init(weights)
    transform_fn = variant(tx.update)
    new_updates, _ = transform_fn(updates, state, weights)
    # Assert output as expected.
    chex.assert_trees_all_close(new_updates, expected_tx_updates)


def test_ema() -> None:
    values = jnp.array([5.0, 7.0])
    decay = 0.9
    d = decay

    ema = Ema[Any](decay=decay, debias=False)
    state = ema.init(values[0])  # init to zeroes

    transform_fn = variant(ema.update)
    mean, state = transform_fn(values[0], state)
    np.testing.assert_allclose(mean, (1 - d) * values[0], atol=1e-4)

    mean, state = transform_fn(values[1], state)
    np.testing.assert_allclose(mean, (1 - d) * (values[1] + d * values[0]), atol=1e-2)


def test_ema_debias() -> None:
    values = jnp.array([5.0, 7.0])
    decay = 0.9
    d = decay

    ema = Ema[Any](decay=decay)
    state = ema.init(values[0])

    transform_fn = variant(ema.update)
    mean, state = transform_fn(values[0], state)
    np.testing.assert_allclose(mean, values[0], atol=1e-4)

    mean, state = transform_fn(values[1], state)
    np.testing.assert_allclose(
        mean, ((1 - d) * values[1] + d * (1 - d) * values[0]) / (1 - d**2), atol=1e-2
    )
    # The state must not be debiased.
    np.testing.assert_allclose(
        state.data.ema, (1 - d) * values[1] + d * (1 - d) * values[0], atol=1e-2
    )


def test_apply_every() -> None:
    # The frequency of the application of SGD
    k = 4
    zero_update = (jnp.array([0.0, 0.0]), jnp.array([0.0, 0.0]))

    # optax SGD
    optax_sgd_params = init_params
    sgd_transform = SGD[tuple[RealArray, RealArray]](LR, 0.0)
    state_sgd = sgd_transform.init(optax_sgd_params)

    # optax SGD plus apply every
    optax_sgd_apply_every_params = init_params
    sgd_apply_every = ChainedGradientTransformation[tuple[RealArray, RealArray]](
        [ApplyEvery[Any](k=k), Trace[Any](decay=0, nesterov=False), Scale[Any](-LR)]
    )
    state_sgd_apply_every = sgd_apply_every.init(optax_sgd_apply_every_params)
    transform_fn = variant(sgd_apply_every.update)

    for i in range(STEPS):
        # Apply a step of SGD
        updates_sgd, state_sgd = sgd_transform.update(per_step_updates, state_sgd, None)
        optax_sgd_params = apply_updates(optax_sgd_params, updates_sgd)

        # Apply a step of sgd_apply_every
        updates_sgd_apply_every, state_sgd_apply_every = transform_fn(
            per_step_updates, state_sgd_apply_every, None
        )
        optax_sgd_apply_every_params = apply_updates(
            optax_sgd_apply_every_params, updates_sgd_apply_every
        )

        # Every k steps, check equivalence.
        if i % k == k - 1:
            chex.assert_trees_all_close(
                optax_sgd_apply_every_params, optax_sgd_params, atol=1e-6, rtol=1e-5
            )
        # Otherwise, check update is zero.
        else:
            chex.assert_trees_all_close(updates_sgd_apply_every, zero_update, atol=0.0, rtol=0.0)


def test_scale() -> None:
    updates = per_step_updates
    for i in range(1, STEPS + 1):
        factor = 0.1**i
        rescaler = Scale[Any](factor)
        empty_state = rescaler.init(updates)
        # Apply rescaling.
        scaled_updates, _ = rescaler.update(updates, empty_state, updates)

        # Manually scale updates.
        def rescale(t: float | Array, *, factor: float = factor) -> float | Array:
            return t * factor

        manual_updates = tree.map(rescale, updates)
        # Check the rescaled updates match.
        chex.assert_trees_all_close(scaled_updates, manual_updates)


@pytest.mark.parametrize(
    ("inputs", "outputs"),
    [
        ([1.0, 2.0], [1.0, 2.0]),
        ([[1.0, 2.0], [3.0, 4.0]], [[-0.5, 0.5], [-0.5, 0.5]]),
        (
            [[[1.0, 2.0], [3.0, 4.0]], [[5.0, 6.0], [7.0, 8.0]]],
            [[[-1.5, -0.5], [0.5, 1.5]], [[-1.5, -0.5], [0.5, 1.5]]],
        ),
    ],
)
def test_centralize(inputs: object, outputs: object) -> None:
    inputs = jnp.asarray(inputs)
    outputs = jnp.asarray(outputs)
    centralizer = Centralize[Any]()
    empty_state = centralizer.init(inputs)
    centralized_inputs, _ = centralizer.update(inputs, empty_state, None)
    chex.assert_trees_all_close(centralized_inputs, outputs)



from __future__ import annotations

from collections.abc import Callable
from typing import Self, override

import jax.numpy as jnp
from jax import tree

from tjax.dataclasses import dataclass

from tjax._src.annotations import PyTree
from tjax._src.math_tools import abs_square


# Gradient states ----------------------------------------------------------------------------------
@dataclass
class GradientState:
    """This base class allows instance checks, and strict type annotations."""


@dataclass
class GenericGradientState(GradientState):
    """A simple container for gradient states.

    Optax doesn't provide a suitable base class.
    """

    data: PyTree

    @classmethod
    def wrap[Weights: PyTree](cls, weights: Weights, state: PyTree) -> tuple[Weights, Self]:
        return weights, cls(state)


# Gradient transformation base classes -------------------------------------------------------------
# The weights type variable must be the same type variable that is used for the gradients.  After
# all, these must be compatible pytrees.
@dataclass
class GradientTransformation[State: GradientState, Weights: PyTree]:
    """A class that transforms weight cotangents into into weight deltas.

    The delta are added to the weights.  The typical goal of a gradient transformation is for the
    weights to converge quickly and stably.

    The most basic gradient transformation is the scaled negative gradients, which can be
    interpreted as doing Newton's method with an assumed isotropic loss Hessian.  In general, all
    gradient transformations can be interpreted as Hessian estimators.
    """

    def init(self, parameters: Weights) -> State:
        raise NotImplementedError

    def update(
        self, gradient: Weights, state: State, parameters: Weights | None
    ) -> tuple[Weights, State]:
        """Transform the weight gradient and update the gradient state.

        Args:
            gradient: The derivative of the loss with respect to the weights.
            state: The gradient state.
            parameters: The weights.

        Returns:
            new_gradient: The modified gradient.
            new_gradient_state: The new gradient state.
        """
        raise NotImplementedError


@dataclass
class SecondOrderGradientTransformation[State: GradientState, Weights: PyTree](
    GradientTransformation[State, Weights]
):
    """A second order gradient transformation.

    This is a special case of a gradient transformation whose update rule accepts a
    Hessian-vector-product function in its update.
    """

    @override
    def update(
        self, gradient: Weights, state: State, parameters: Weights | None
    ) -> tuple[Weights, State]:
        """Transform the weight gradient and update the gradient state.

        Args:
            gradient: The derivative of the loss with respect to the weights.
            state: The gradient state.
            parameters: The weights.

        Returns:
            new_gradient: The modified gradient.
            new_gradient_state: The new gradient state.

        This uses the outer product approximation of the Hessian.
        """

        def hessian_vector_product(v: Weights) -> Weights:
            d = tree.reduce_associative(jnp.add, tree.map(jnp.vdot, gradient, v), identity=0.0)  # type: ignore
            return tree.map(lambda x: x * d, gradient)

        return self.second_order_update(gradient, state, parameters, hessian_vector_product)

    def second_order_update(
        self,
        gradient: Weights,
        state: State,
        parameters: Weights | None,
        hessian_vector_product: Callable[[Weights], Weights],
    ) -> tuple[Weights, State]:
        """Transform the weight gradient and update the gradient state.

        Args:
            gradient: The derivative of the loss with respect to the weights.
            state: The gradient state.
            parameters: The weights.
            hessian_vector_product: A function that maps v to the Hessian of the loss with respect
                to the weights times v.

        Returns:
            new_gradient: The modified gradient.
            new_gradient_state: The new gradient state.
        """
        raise NotImplementedError


@dataclass
class ThirdOrderGradientTransformation[State: GradientState, Weights: PyTree](
    SecondOrderGradientTransformation[State, Weights]
):
    """A third order gradient transformation.

    This is a special case of a second order gradient transformation whose update rule accepts the
    diagonal entries of the Hessian.
    """

    @override
    def second_order_update(
        self,
        gradient: Weights,
        state: State,
        parameters: Weights | None,
        hessian_vector_product: Callable[[Weights], Weights],
    ) -> tuple[Weights, State]:
        """Transform the weight gradient and update the gradient state.

        Args:
            gradient: The derivative of the loss with respect to the weights.
            state: The gradient state.
            parameters: The weights.
            hessian_vector_product: A function that maps v to the Hessian of the loss with respect
                to the weights times v.

        Returns:
            new_gradient: The modified gradient.
            new_gradient_state: The new gradient state.

        Uses the outer product approximation of the Hessian to provide the diagonal entries of the
        Hessian.
        """
        return self.third_order_update(
            gradient, state, parameters, hessian_vector_product, tree.map(abs_square, gradient)
        )

    def third_order_update(
        self,
        gradient: Weights,
        state: State,
        parameters: Weights | None,
        hessian_vector_product: Callable[[Weights], Weights],
        hessian_diagonal: Weights,
    ) -> tuple[Weights, State]:
        """Transform the weight gradient and update the gradient state.

        Args:
            gradient: The derivative of the loss with respect to the weights.
            state: The gradient state.
            parameters: The weights.
            hessian_vector_product: A function that maps v to the Hessian of the loss with respect
                to the weights times v.
            hessian_diagonal: The diagonal entries of the Hessian of the loss with respect to the
                weights.

        Returns:
            new_gradient: The modified gradient.
            new_gradient_state: The new gradient state.
        """
        raise NotImplementedError



import math
from functools import partial

import jax
import jax.numpy as jnp
import pytest
import scipy.special as sc
from jax import vjp
from numpy.testing import assert_allclose

from tjax import Array, JaxRealArray, bessel_iv_ratio, divide_where, log_bessel_ive, normalize


@pytest.mark.parametrize(
    ("x", "axis", "result"),
    [
        ((1.0, 2.0, 3.0, 4.0), 0, (0.1, 0.2, 0.3, 0.4)),
        ((0.0, 0.0), 0, (0.5, 0.5)),
    ],
)
def test_l1(
    x: object,
    axis: tuple[int, ...] | int | None,
    result: JaxRealArray,
) -> None:
    assert_allclose(normalize("l1", jnp.asarray(x), axis=axis), jnp.asarray(result))


@pytest.mark.parametrize(
    ("x", "axis", "result"),
    [
        ((5.0, 12.0), 0, (5 / 13, 12 / 13)),
        ((0.0, 0.0), 0, jnp.ones(2) / jnp.sqrt(2)),
    ],
)
def test_l2(
    x: object,
    axis: tuple[int, ...] | int | None,
    result: JaxRealArray,
) -> None:
    assert_allclose(normalize("l2", jnp.asarray(x), axis=axis), jnp.asarray(result))


@pytest.mark.parametrize("k", [0.0, 1.0])
def test_divide_where(k: float) -> None:
    s = (5, 3)
    w = jnp.ones(s) * k
    x = jnp.arange(math.prod(w.shape), dtype="f").reshape(w.shape)
    dummy = jnp.ones_like(w[..., 0])

    def f(w: Array, x: Array, dummy: Array) -> Array:
        total_w = jnp.sum(w, axis=-1)
        return divide_where(
            dividend=jnp.sum(w * x, axis=-1),  # type: ignore
            divisor=total_w,
            where=total_w > 0.0,
            otherwise=dummy,
        )

    y, vjp_f = vjp(partial(f, dummy=dummy), w, x)
    w_bar, x_bar = vjp_f(jnp.ones_like(y))
    assert_allclose(w_bar, k / s[-1] * jnp.tile(jnp.asarray([-1, 0, 1]), (*s[:-1], 1)), atol=1e-3)
    assert_allclose(x_bar, k / s[-1] * jnp.ones(s))


@pytest.mark.parametrize(
    ("v", "x"),
    [(0.0, 0.1), (0.5, 2.0), (2.3, 3.0), (10.0, 5.0)],
)
def test_bessel_iv_ratio_value(v: float, x: float) -> None:
    expected = sc.ive(v + 1.0, x) / sc.ive(v, x)
    assert_allclose(bessel_iv_ratio(jnp.asarray(v), jnp.asarray(x)), expected, rtol=1e-6)


@pytest.mark.parametrize(
    ("v", "x"),
    [(0.5, 0.8), (2.3, 1.5)],
)
def test_bessel_iv_ratio_derivatives(v: float, x: float) -> None:
    step = 1e-4
    expected_dv = (
        sc.ive(v + step + 1.0, x) / sc.ive(v + step, x)
        - sc.ive(v - step + 1.0, x) / sc.ive(v - step, x)
    ) / (2.0 * step)
    expected_dx = (
        sc.ive(v + 1.0, x + step) / sc.ive(v, x + step)
        - sc.ive(v + 1.0, x - step) / sc.ive(v, x - step)
    ) / (2.0 * step)
    actual_dv, actual_dx = jax.jacfwd(bessel_iv_ratio, argnums=(0, 1))(
        jnp.asarray(v), jnp.asarray(x)
    )
    assert_allclose(actual_dv, expected_dv, rtol=2e-4, atol=2e-5)
    assert_allclose(actual_dx, expected_dx, rtol=2e-4, atol=2e-5)


@pytest.mark.parametrize(
    ("v", "x"),
    [(0.0, 0.1), (0.5, 2.0), (2.3, 1.5)],
)
def test_log_bessel_ive_value(v: float, x: float) -> None:
    expected = math.log(sc.ive(v, x))
    assert_allclose(log_bessel_ive(jnp.asarray(v), jnp.asarray(x)), expected, rtol=1e-6)


@pytest.mark.parametrize(
    ("v", "x"),
    [(0.5, 0.8), (2.3, 1.5)],
)
def test_log_bessel_ive_derivatives(v: float, x: float) -> None:
    step = 1e-4
    expected_dv = (math.log(sc.ive(v + step, x)) - math.log(sc.ive(v - step, x))) / (2.0 * step)
    expected_dx = (math.log(sc.ive(v, x + step)) - math.log(sc.ive(v, x - step))) / (2.0 * step)
    actual_dv, actual_dx = jax.jacfwd(log_bessel_ive, argnums=(0, 1))(
        jnp.asarray(v), jnp.asarray(x)
    )
    assert_allclose(actual_dv, expected_dv, rtol=5e-3, atol=5e-4)
    assert_allclose(actual_dx, expected_dx, rtol=2e-5, atol=2e-6)



from __future__ import annotations

from collections.abc import Callable
from typing import Any, override

from optax import (
    add_decayed_weights,
    add_noise,
    apply_every,
    centralize,
    ema,
    scale,
    scale_by_adam,
    scale_by_belief,
    scale_by_param_block_norm,
    scale_by_param_block_rms,
    scale_by_radam,
    scale_by_rms,
    scale_by_rss,
    scale_by_schedule,
    scale_by_sm3,
    scale_by_stddev,
    scale_by_trust_ratio,
    scale_by_yogi,
    trace,
)

from tjax.dataclasses import as_shallow_dict, dataclass, field

from tjax._src.annotations import IntegralNumeric, KeyArray, PyTree, RealNumeric
from .transform import GenericGradientState, GradientTransformation


# New classes --------------------------------------------------------------------------------------
@dataclass
class Schedule:
    """The schedule of step sizes.

    This class differs from optax.Schedule in that it's a pytree.  It can therefore be passed
    dynamically (since it marks the callable as static).  This allows it to participate in the union
    ScalarOrSchedule.
    """

    step_size_fn: Callable[[IntegralNumeric], RealNumeric] = field(static=True)

    def __call__(self, count: IntegralNumeric) -> RealNumeric:
        return self.step_size_fn(count)


# Transforms from optax._src.transform.py ----------------------------------------------------------
@dataclass
class Trace[Weights: PyTree](GradientTransformation[GenericGradientState, Weights]):
    """Compute a trace of past updates.

    Note: `trace` and `ema` have very similar but distinct updates;
    `trace = decay * trace + t`, while `ema = decay * ema + (1-decay) * t`.
    Both are frequently found in the optimisation literature.

    Args:
        decay: the decay rate for the trace of past updates.
        nesterov: whether to use Nesterov momentum.
        accumulator_dtype: optional `dtype` to be used for the accumulator; if
            `None` then the `dtype` is inferred from `params` and `updates`.
    """

    decay: RealNumeric
    nesterov: bool = field(default=False, static=True)
    accumulator_dtype: Any = field(default=None, static=True)

    @override
    def init(self, parameters: Weights) -> GenericGradientState:
        return GenericGradientState(
            trace(
                self.decay,
                self.nesterov,
                self.accumulator_dtype,  # pyright: ignore
            ).init(parameters)
        )

    @override
    def update(
        self, gradient: Weights, state: GenericGradientState, parameters: Weights | None
    ) -> tuple[Weights, GenericGradientState]:
        t = trace(self.decay, self.nesterov, self.accumulator_dtype)  # pyright: ignore
        return GenericGradientState.wrap(  # pyright: ignore # type: ignore
            *t.update(gradient, state.data, parameters)
        )


@dataclass
class Ema[Weights: PyTree](GradientTransformation[GenericGradientState, Weights]):
    """Compute an exponential moving average of past updates.

    Note: `trace` and `ema` have very similar but distinct updates;
    `ema = decay * ema + (1-decay) * t`, while `trace = decay * trace + t`.
    Both are frequently found in the optimisation literature.

    Args:
        decay: the decay rate for the exponential moving average.
        debias: whether to debias the transformed gradient.
        accumulator_dtype: optional `dtype` to used for the accumulator; if `None`
            then the `dtype` is inferred from `params` and `updates`.
    """

    decay: RealNumeric
    debias: bool = field(default=True, static=True)
    accumulator_dtype: Any = field(default=None, static=True)

    @override
    def init(self, parameters: Weights) -> GenericGradientState:
        t = ema(self.decay, self.debias, self.accumulator_dtype)  # pyright: ignore
        return GenericGradientState(t.init(parameters))

    @override
    def update(
        self, gradient: Weights, state: GenericGradientState, parameters: Weights | None = None
    ) -> tuple[Weights, GenericGradientState]:
        return GenericGradientState.wrap(  # pyright: ignore # type: ignore
            *ema(**as_shallow_dict(self)).update(gradient, state.data, parameters)
        )


@dataclass
class ScaleByRss[Weights: PyTree](GradientTransformation[GenericGradientState, Weights]):
    """Rescale updates by the root of the sum of all squared gradients to date.

    References:
        [Duchi et al, 2011](https://jmlr.org/papers/volume12/duchi11a/duchi11a.pdf)
        [McMahan et al., 2010](https://arxiv.org/abs/1002.4908)

    Args:
        initial_accumulator_value: Starting value for accumulators, must be >= 0.
        eps: A small floating point value to avoid zero denominator.
    """

    initial_accumulator_value: RealNumeric = 0.1
    eps: RealNumeric = 1e-7

    @override
    def init(self, parameters: Weights) -> GenericGradientState:
        t = scale_by_rss(self.initial_accumulator_value, self.eps)  # pyright: ignore
        return GenericGradientState(t.init(parameters))

    @override
    def update(
        self, gradient: Weights, state: GenericGradientState, parameters: Weights | None = None
    ) -> tuple[Weights, GenericGradientState]:
        t = scale_by_rss(self.initial_accumulator_value, self.eps)  # pyright: ignore
        return GenericGradientState.wrap(  # pyright: ignore # type: ignore
            *t.update(gradient, state.data, parameters)
        )


@dataclass
class ScaleByRms[Weights: PyTree](GradientTransformation[GenericGradientState, Weights]):
    """Rescale updates by the root of the exp. moving avg of the square.

    References:
        [Hinton](www.cs.toronto.edu/~tijmen/csc321/slides/lecture_slides_lec6.pdf)

    Args:
        decay: decay rate for the exponentially weighted average of squared grads.
        eps: term added to the denominator to improve numerical stability.
        initial_scale: initial value for second moment
    """

    decay: RealNumeric = 0.9
    eps: RealNumeric = 1e-8
    initial_scale: RealNumeric = 0.0

    @override
    def init(self, parameters: Weights) -> GenericGradientState:
        t = scale_by_rms(self.decay, self.eps, self.initial_scale)  # pyright: ignore
        return GenericGradientState(t.init(parameters))

    @override
    def update(
        self, gradient: Weights, state: GenericGradientState, parameters: Weights | None
    ) -> tuple[Weights, GenericGradientState]:
        t = scale_by_rms(self.decay, self.eps, self.initial_scale)  # pyright: ignore
        return GenericGradientState.wrap(  # pyright: ignore # type: ignore
            *t.update(gradient, state.data, parameters)
        )


@dataclass
class ScaleByStddev[Weights: PyTree](GradientTransformation[GenericGradientState, Weights]):
    """Rescale updates by the root of the centered exp. moving average of squares.

    References:
        [Hinton](www.cs.toronto.edu/~tijmen/csc321/slides/lecture_slides_lec6.pdf)

    Args:
        decay: decay rate for the exponentially weighted average of squared grads.
        eps: term added to the denominator to improve numerical stability.
        initial_scale: initial value for second moment
    """

    decay: RealNumeric = 0.9
    eps: RealNumeric = 1e-8
    initial_scale: RealNumeric = 0.0

    @override
    def init(self, parameters: Weights) -> GenericGradientState:
        t = scale_by_stddev(self.decay, self.eps, self.initial_scale)  # pyright: ignore
        return GenericGradientState(t.init(parameters))

    @override
    def update(
        self, gradient: Weights, state: GenericGradientState, parameters: Weights | None
    ) -> tuple[Weights, GenericGradientState]:
        t = scale_by_stddev(self.decay, self.eps, self.initial_scale)  # pyright: ignore
        return GenericGradientState.wrap(  # pyright: ignore # type: ignore
            *t.update(gradient, state.data, parameters)
        )


@dataclass
class ScaleByAdam[Weights: PyTree](GradientTransformation[GenericGradientState, Weights]):
    b1: RealNumeric = 0.9
    b2: RealNumeric = 0.999
    eps: RealNumeric = 1e-8
    eps_root: RealNumeric = 0.0
    mu_dtype: Any = field(default=None, static=True)

    @override
    def init(self, parameters: Weights) -> GenericGradientState:
        return GenericGradientState(scale_by_adam(**as_shallow_dict(self)).init(parameters))

    @override
    def update(
        self, gradient: Weights, state: GenericGradientState, parameters: Weights | None
    ) -> tuple[Weights, GenericGradientState]:
        return GenericGradientState.wrap(  # pyright: ignore # type: ignore
            *scale_by_adam(**as_shallow_dict(self)).update(gradient, state.data, parameters)
        )


@dataclass
class Scale[Weights: PyTree](GradientTransformation[GenericGradientState, Weights]):
    """Scale updates by some fixed scalar `step_size`.

    Args:
        step_size: a scalar corresponding to a fixed scaling factor for updates.
    """

    step_size: RealNumeric

    @override
    def init(self, parameters: Weights) -> GenericGradientState:
        return GenericGradientState(scale(**as_shallow_dict(self)).init(parameters))

    @override
    def update(
        self, gradient: Weights, state: GenericGradientState, parameters: Weights | None
    ) -> tuple[Weights, GenericGradientState]:
        return GenericGradientState.wrap(  # pyright: ignore # type: ignore
            *scale(**as_shallow_dict(self)).update(gradient, state.data, parameters)
        )


@dataclass
class ScaleByParamBlockNorm[Weights: PyTree](GradientTransformation[GenericGradientState, Weights]):
    """Scale updates for each param block by the norm of that block's parameters.

    A `block` is here a weight vector (e.g. in a Linear layer) or a weight matrix
    (e.g. in a convolutional layer) appearing as a leaf in the grads/param pytree.

    Args:
        min_scale: minimum scaling factor.
    """

    min_scale: RealNumeric = 1e-3

    @override
    def init(self, parameters: Weights) -> GenericGradientState:
        t = scale_by_param_block_norm(self.min_scale)  # pyright: ignore
        return GenericGradientState(t.init(parameters))

    @override
    def update(
        self, gradient: Weights, state: GenericGradientState, parameters: Weights | None
    ) -> tuple[Weights, GenericGradientState]:
        t = scale_by_param_block_norm(self.min_scale)  # pyright: ignore
        return GenericGradientState.wrap(  # pyright: ignore
            *t.update(gradient, state.data, parameters)
        )  # type: ignore


@dataclass
class ScaleByParamBlockRMS[Weights: PyTree](GradientTransformation[GenericGradientState, Weights]):
    """Scale updates by rms of the gradient for each param vector or matrix.

    A `block` is here a weight vector (e.g. in a Linear layer) or a weight matrix
    (e.g. in a convolutional layer) appearing as a leaf in the grads/param pytree.

    Args:
        min_scale: minimum scaling factor.
    """

    min_scale: RealNumeric = 1e-3

    @override
    def init(self, parameters: Weights) -> GenericGradientState:
        t = scale_by_param_block_rms(self.min_scale).init(parameters)  # pyright: ignore
        return GenericGradientState(t)

    @override
    def update(
        self, gradient: Weights, state: GenericGradientState, parameters: Weights | None
    ) -> tuple[Weights, GenericGradientState]:
        t = scale_by_param_block_rms(self.min_scale)  # pyright: ignore
        return GenericGradientState.wrap(  # pyright: ignore # type: ignore
            *t.update(gradient, state.data, parameters)
        )


@dataclass
class ScaleByBelief[Weights: PyTree](GradientTransformation[GenericGradientState, Weights]):
    """Rescale updates according to the AdaBelief algorithm.

    References:
      [Zhuang et al, 2020](https://arxiv.org/abs/2010.07468)

    Args:
      b1: decay rate for the exponentially weighted average of grads.
      b2: decay rate for the exponentially weighted average of variance of grads.
      eps: term added to the denominator to improve numerical stability.
      eps_root: term added to the denominator inside the square-root to improve
        numerical stability when backpropagating gradients through the rescaling.
    """

    b1: RealNumeric = 0.9
    b2: RealNumeric = 0.999
    eps: RealNumeric = 0.0
    eps_root: RealNumeric = 1e-16

    @override
    def init(self, parameters: Weights) -> GenericGradientState:
        t = scale_by_belief(self.b1, self.b2, self.eps, self.eps_root)  # pyright: ignore
        return GenericGradientState(t.init(parameters))

    @override
    def update(
        self, gradient: Weights, state: GenericGradientState, parameters: Weights | None
    ) -> tuple[Weights, GenericGradientState]:
        t = scale_by_belief(self.b1, self.b2, self.eps, self.eps_root)  # pyright: ignore
        return GenericGradientState.wrap(  # pyright: ignore # type: ignore
            *t.update(gradient, state.data, parameters)
        )


@dataclass
class ScaleByYogi[Weights: PyTree](GradientTransformation[GenericGradientState, Weights]):
    """Rescale updates according to the Yogi algorithm.

    References:
        [Zaheer et al, 2018](https://papers.nips.cc/paper/2018/hash/90365351ccc7437a1309dc64e4db32a3-Abstract.html)

    Args:
        b1: decay rate for the exponentially weighted average of grads.
        b2: decay rate for the exponentially weighted average of variance of grads.
        eps: term added to the denominator to improve numerical stability.
        eps_root: term added to the denominator inside the square-root to improve
            numerical stability when backpropagating gradients through the rescaling.
        initial_accumulator_value: The starting value for accumulators.
            Only positive values are allowed.
    """

    b1: RealNumeric = 0.9
    b2: RealNumeric = 0.999
    eps: RealNumeric = 1e-3
    eps_root: RealNumeric = 0.0
    initial_accumulator_value: RealNumeric = 1e-6

    @override
    def init(self, parameters: Weights) -> GenericGradientState:
        t = scale_by_yogi(self.b1, self.b2, self.eps, self.eps_root)  # pyright: ignore
        return GenericGradientState(t.init(parameters))

    @override
    def update(
        self, gradient: Weights, state: GenericGradientState, parameters: Weights | None
    ) -> tuple[Weights, GenericGradientState]:
        t = scale_by_yogi(self.b1, self.b2, self.eps, self.eps_root)  # pyright: ignore
        return GenericGradientState.wrap(  # pyright: ignore # type: ignore
            *t.update(gradient, state.data, parameters)
        )


@dataclass
class ScaleByRAdam[Weights: PyTree](GradientTransformation[GenericGradientState, Weights]):
    """Rescale updates according to the Rectified Adam algorithm.

    References:
        [Liu et al, 2020](https://arxiv.org/abs/1908.03265)

    Args:
        b1: decay rate for the exponentially weighted average of grads.
        b2: decay rate for the exponentially weighted average of squared grads.
        eps: term added to the denominator to improve numerical stability.
        eps_root: term added to the denominator inside the square-root to improve
            numerical stability when backpropagating gradients through the rescaling.
        threshold: Threshold for variance tractability
    """

    b1: RealNumeric = 0.9
    b2: RealNumeric = 0.999
    eps: RealNumeric = 1e-8
    eps_root: RealNumeric = 0.0
    threshold: RealNumeric = 5.0

    @override
    def init(self, parameters: Weights) -> GenericGradientState:
        t = scale_by_radam(self.b1, self.b2, self.eps, self.eps_root)  # pyright: ignore
        return GenericGradientState(t.init(parameters))

    @override
    def update(
        self, gradient: Weights, state: GenericGradientState, parameters: Weights | None
    ) -> tuple[Weights, GenericGradientState]:
        t = scale_by_radam(self.b1, self.b2, self.eps, self.eps_root)  # pyright: ignore
        return GenericGradientState.wrap(  # pyright: ignore # type: ignore
            *t.update(gradient, state.data, parameters)
        )


@dataclass
class AddDecayedWeights[Weights: PyTree](GradientTransformation[GenericGradientState, Weights]):
    """Add parameter scaled by `weight_decay`.

    Args:
        weight_decay: a scalar weight decay rate.
        mask: a tree with same structure as (or a prefix of) the params PyTree,
            or a Callable that returns such a pytree given the params/updates.
            The leaves should be booleans, `True` for leaves/subtrees you want to
            apply the transformation to, and `False` for those you want to skip.
        weight_decay: RealNumeric = 0.0
    """

    weight_decay: RealNumeric = 0.0
    mask: bool | Weights | Callable[[Weights], Any] | None = field(default=None, static=True)

    @override
    def init(self, parameters: Weights) -> GenericGradientState:
        return GenericGradientState(add_decayed_weights(**as_shallow_dict(self)).init(parameters))

    @override
    def update(
        self, gradient: Weights, state: GenericGradientState, parameters: Weights | None
    ) -> tuple[Weights, GenericGradientState]:
        return GenericGradientState.wrap(  # pyright: ignore # type: ignore
            *add_decayed_weights(**as_shallow_dict(self)).update(gradient, state.data, parameters)
        )


@dataclass
class ScaleBySchedule[Weights: PyTree](GradientTransformation[GenericGradientState, Weights]):
    """Scale updates using a custom schedule for the `step_size`.

    Args:
        step_size_fn: a function that takes an update count as input and proposes
            the step_size to multiply the updates by.
    """

    step_size_fn: Schedule

    @override
    def init(self, parameters: Weights) -> GenericGradientState:
        t = scale_by_schedule(self.step_size_fn)  # pyright: ignore
        return GenericGradientState(t.init(parameters))

    @override
    def update(
        self, gradient: Weights, state: GenericGradientState, parameters: Weights | None
    ) -> tuple[Weights, GenericGradientState]:
        t = scale_by_schedule(self.step_size_fn)  # pyright: ignore
        return GenericGradientState.wrap(  # pyright: ignore # type: ignore
            *t.update(gradient, state.data, parameters)
        )


@dataclass
class ScaleByTrustRatio[Weights: PyTree](GradientTransformation[GenericGradientState, Weights]):
    """Scale updates by trust ratio`.

    References:
        [You et. al 2020](https://arxiv.org/abs/1904.00962)

    Args:
        min_norm: minimum norm for params and gradient norms; by default is zero.
        trust_coefficient: a multiplier for the trust ratio.
        eps: additive constant added to the denominator for numerical stability.
    """

    min_norm: RealNumeric = 0.0
    trust_coefficient: RealNumeric = 1.0
    eps: RealNumeric = 0.0

    @override
    def init(self, parameters: Weights) -> GenericGradientState:
        return GenericGradientState(scale_by_trust_ratio(**as_shallow_dict(self)).init(parameters))

    @override
    def update(
        self, gradient: Weights, state: GenericGradientState, parameters: Weights | None
    ) -> tuple[Weights, GenericGradientState]:
        return GenericGradientState.wrap(  # pyright: ignore # type: ignore
            *scale_by_trust_ratio(**as_shallow_dict(self)).update(gradient, state.data, parameters)
        )


@dataclass
class AddNoise[Weights: PyTree](GradientTransformation[GenericGradientState, Weights]):
    """Add gradient noise.

    References:
        [Neelakantan et al, 2014](https://arxiv.org/abs/1511.06807)

    Args:
        eta: base variance of the gaussian noise added to the gradient.
        gamma: decay exponent for annealing of the variance.
        seed: seed for random number generation.
    """

    eta: RealNumeric
    gamma: RealNumeric
    rng: KeyArray

    @override
    def init(self, parameters: Weights) -> GenericGradientState:
        return GenericGradientState(add_noise(**as_shallow_dict(self)).init(parameters))

    @override
    def update(
        self, gradient: Weights, state: GenericGradientState, parameters: Weights | None
    ) -> tuple[Weights, GenericGradientState]:
        return GenericGradientState.wrap(  # pyright: ignore # type: ignore
            *add_noise(**as_shallow_dict(self)).update(gradient, state.data, parameters)
        )


@dataclass
class ApplyEvery[Weights: PyTree](GradientTransformation[GenericGradientState, Weights]):
    """Accumulate gradients and apply them every k steps.

    Note that if this transformation is part of a chain, the states of the other
    transformations will still be updated at every step. In particular, using
    `apply_every` with a batch size of N/2 and k=2 is not necessarily equivalent
    to not using `apply_every` with a batch size of N. If this equivalence is
    important for you, consider using the `optax.MultiSteps`.

    Args:
        k: emit non-zero gradients every k steps, otherwise accumulate them.
    """

    k: int = field(default=1, static=True)

    @override
    def init(self, parameters: Weights) -> GenericGradientState:
        return GenericGradientState(apply_every(self.k).init(parameters))

    @override
    def update(
        self, gradient: Weights, state: GenericGradientState, parameters: Weights | None
    ) -> tuple[Weights, GenericGradientState]:
        return GenericGradientState.wrap(  # pyright: ignore # type: ignore
            *apply_every(self.k).update(gradient, state.data, parameters)
        )


@dataclass
class Centralize[Weights: PyTree](GradientTransformation[GenericGradientState, Weights]):
    """Centralize gradients.

    References:
        [Yong et al, 2020](https://arxiv.org/abs/2004.01461)
    """

    @override
    def init(self, parameters: Weights) -> GenericGradientState:
        return GenericGradientState(centralize().init(parameters))

    @override
    def update(
        self, gradient: Weights, state: GenericGradientState, parameters: Weights | None
    ) -> tuple[Weights, GenericGradientState]:
        return GenericGradientState.wrap(  # pyright: ignore # type: ignore
            *centralize().update(gradient, state.data, parameters)
        )


@dataclass
class ScaleBySM3[Weights: PyTree](GradientTransformation[GenericGradientState, Weights]):
    """Scale updates by sm3`.

    References:
        [Anil et. al 2019](https://arxiv.org/abs/1901.11150)

    Args:
        b1: decay rate for the exponentially weighted average of grads.
        b2: decay rate for the exponentially weighted average of squared grads.
        eps: term added to the denominator to improve numerical stability.
    """

    b1: RealNumeric = 0.9
    b2: RealNumeric = 1.0
    eps: RealNumeric = 1e-8

    @override
    def init(self, parameters: Weights) -> GenericGradientState:
        return GenericGradientState(scale_by_sm3(self.b1, self.b2, self.eps).init(parameters))  # pyright: ignore

    @override
    def update(
        self, gradient: Weights, state: GenericGradientState, parameters: Weights | None
    ) -> tuple[Weights, GenericGradientState]:
        t = scale_by_sm3(self.b1, self.b2, self.eps)  # pyright: ignore
        return GenericGradientState.wrap(  # pyright: ignore # type: ignore
            *t.update(gradient, state.data, parameters)
        )



from __future__ import annotations

from collections.abc import Sequence
from types import ModuleType
from typing import Any

import numpy as np
import optype.numpy as onp
import optype.numpy.compat as npc
from jax import Array as JaxArray

type Shape = tuple[int, ...]
type ShapeLike = int | Sequence[int]
type _SliceLikeItem = int | slice | None
type SliceLike = _SliceLikeItem | tuple[_SliceLikeItem, ...]


type JaxBooleanArray = JaxArray
type JaxIntegralArray = JaxArray
type JaxRealArray = JaxArray
type JaxComplexArray = JaxArray
type KeyArray = JaxArray


type NumpyArray = onp.ArrayND
type NumpyBooleanArray = onp.ArrayND[np.bool_]
type NumpyIntegralArray = onp.ArrayND[npc.integer]
type NumpyRealArray = onp.ArrayND[npc.floating]
type NumpyComplexArray = onp.ArrayND[npc.number]
type BooleanArray = NumpyBooleanArray | JaxArray
type IntegralArray = NumpyIntegralArray | JaxArray
type RealArray = NumpyRealArray | JaxArray
type ComplexArray = NumpyComplexArray | JaxArray


type NumpyBooleanNumeric = NumpyBooleanArray | bool
type NumpyIntegralNumeric = NumpyIntegralArray | int
type NumpyRealNumeric = NumpyRealArray | float
type NumpyComplexNumeric = NumpyComplexArray | complex
type BooleanNumeric = BooleanArray | bool
type IntegralNumeric = IntegralArray | int
type RealNumeric = RealArray | float
type ComplexNumeric = ComplexArray | complex


# Eventually, these will come from array_api_extra.
type Array = NumpyArray | JaxArray
type Device = Any
type DType = Any
type Namespace = ModuleType


type PyTree = Any



from __future__ import annotations

import re

import pytest

from tjax import assert_tree_allclose, tree_allclose
from tjax.dataclasses import dataclass


@pytest.mark.parametrize(
    ("actual", "desired"),
    [
        (0.1234567890123, "0\\.123457"),
        (0.000123456789, "0\\.00012346"),
        (1.23456789e-8, "1\\.e-08"),
        (1.23456789e8, "1\\.23457e\\+08"),
    ],
)
def test_testing(capsys: pytest.CaptureFixture[str], actual: float, desired: str) -> None:
    with pytest.raises(AssertionError) as exception:
        assert_tree_allclose(actual, 1e6, rtol=1e-5, atol=1e-8)
    captured = capsys.readouterr()
    assert not captured.out  # Nothing should be printed.
    assert not captured.err
    assertion_string = exception.value.args[0]
    pattern = rf"""
Tree leaves don't match at position 0 with rtol=1e-05 and atol=1e-08\.
Mismatched elements: 1 / 1 \(100%\)
Maximum absolute difference among violations: .+
Maximum relative difference among violations: .+

Actual: {actual}
Desired: 1000000\.0
Test string:
desired = {desired}"""
    assert re.match(pattern, assertion_string)


@pytest.mark.parametrize(
    ("actual", "desired", "result"),
    [
        ((), (), True),
        ((1.2, 3.4), (1.2, 3.4), True),
        ({"a": 1.2}, {"a": 1.2}, True),
        (1.3, 1.4, False),
    ],
)
def test_tree_allclose(actual: object, desired: object, *, result: bool | None) -> None:
    assert tree_allclose(actual, desired) == result


def test_tree_allclose_structure_mismatch() -> None:
    with pytest.raises(ValueError, match="Tuple arity mismatch"):
        tree_allclose((1, 2), (1, 2, 3))


def test_assert_tree_allclose_relative_dataclass_message() -> None:
    @dataclass
    class Point:
        x: int
        y: int

    original = Point(1, 2)
    desired = Point(0, 0)
    actual = Point(3, 4)

    with pytest.raises(AssertionError) as exception:
        assert_tree_allclose(actual, desired, "original", original)

    assertion_string = exception.value.args[0]
    assert "desired = replace(original, x=3,\ny=4)" in assertion_string



from __future__ import annotations

from typing import Any, override

from tjax.dataclasses import dataclass

from tjax._src.annotations import PyTree
from .transform import GradientState, GradientTransformation


@dataclass
class ChainedGradientState(GradientState):
    sub_states: list[PyTree]


@dataclass
class ChainedGradientTransformation[Weights: PyTree](
    GradientTransformation[ChainedGradientState, Weights]
):
    transforms: list[GradientTransformation[Any, Weights]]

    @override
    def init(self, parameters: Weights) -> ChainedGradientState:
        return ChainedGradientState([transform.init(parameters) for transform in self.transforms])

    @override
    def update(
        self, gradient: Weights, state: ChainedGradientState, parameters: Weights | None
    ) -> tuple[Weights, ChainedGradientState]:
        new_state: list[PyTree] = []
        for sub_state, transform in zip(state.sub_states, self.transforms, strict=True):
            gradient, new_sub_state = transform.update(gradient, sub_state, parameters)
            new_state.append(new_sub_state)
        return gradient, ChainedGradientState(new_state)



from tjax._src.gradient.aliases import (
    DPSGD,
    LARS,
    LBFGS,
    SGD,
    SM3,
    AdaBelief,
    AdaDelta,
    AdaFactor,
    AdaGrad,
    Adam,
    Adamax,
    AdamaxW,
    AdamW,
    Fromage,
    Lamb,
    Lion,
    NoisySGD,
    Novograd,
    OptimisticGradientDescent,
    PolyakSGD,
    RAdam,
    RMSProp,
    Yogi,
)
from tjax._src.gradient.chain import ChainedGradientState, ChainedGradientTransformation
from tjax._src.gradient.smd import SMDGradient, SMDState
from tjax._src.gradient.transform import (
    GenericGradientState,
    GradientState,
    GradientTransformation,
    SecondOrderGradientTransformation,
    ThirdOrderGradientTransformation,
)
from tjax._src.gradient.transforms import (
    AddDecayedWeights,
    AddNoise,
    ApplyEvery,
    Centralize,
    Ema,
    Scale,
    ScaleByAdam,
    ScaleByBelief,
    ScaleByParamBlockNorm,
    ScaleByParamBlockRMS,
    ScaleByRAdam,
    ScaleByRms,
    ScaleByRss,
    ScaleBySchedule,
    ScaleBySM3,
    ScaleByStddev,
    ScaleByTrustRatio,
    ScaleByYogi,
    Schedule,
    Trace,
)

__all__ = [
    "DPSGD",
    "LARS",
    "LBFGS",
    "SGD",
    "SM3",
    "AdaBelief",
    "AdaDelta",
    "AdaFactor",
    "AdaGrad",
    "Adam",
    "AdamW",
    "Adamax",
    "AdamaxW",
    "AddDecayedWeights",
    "AddNoise",
    "ApplyEvery",
    "Centralize",
    "ChainedGradientState",
    "ChainedGradientTransformation",
    "Ema",
    "Fromage",
    "GenericGradientState",
    "GradientState",
    "GradientTransformation",
    "Lamb",
    "Lion",
    "NoisySGD",
    "Novograd",
    "OptimisticGradientDescent",
    "PolyakSGD",
    "RAdam",
    "RMSProp",
    "SMDGradient",
    "SMDState",
    "Scale",
    "ScaleByAdam",
    "ScaleByBelief",
    "ScaleByParamBlockNorm",
    "ScaleByParamBlockRMS",
    "ScaleByRAdam",
    "ScaleByRms",
    "ScaleByRss",
    "ScaleBySM3",
    "ScaleBySchedule",
    "ScaleByStddev",
    "ScaleByTrustRatio",
    "ScaleByYogi",
    "Schedule",
    "SecondOrderGradientTransformation",
    "ThirdOrderGradientTransformation",
    "Trace",
    "Yogi",
]



from __future__ import annotations

from typing import Literal

import jax
import jax.numpy as jnp
import numpy as np
import scipy.special as sc
from array_api_compat import array_namespace, is_jax_namespace, is_torch_namespace

from .annotations import Array, BooleanArray, Namespace


def abs_square[T: Array](x: T) -> T:
    xp = array_namespace(x)
    return xp.square(x.real) + xp.square(x.imag)  # pyright: ignore


# TODO: Remove when it's added to theArrayAPI:
# https://github.com/data-apis/array-api/issues/242
def outer_product[T: Array](x: T, y: T) -> T:
    """Return the broadcasted outer product of a vector with itself.

    This is xp.einsum("...i,...j->...ij", x, y).
    """
    xp = array_namespace(x, y)
    xi = xp.reshape(x, (*x.shape, 1))
    yj = xp.reshape(y.conjugate(), (*y.shape[:-1], 1, y.shape[-1]))
    return xi * yj


def matrix_vector_mul[T: Array](x: T, y: T) -> T:
    """Return the matrix-vector product.

    This is xp.einsum("...ij,...j->...i", x, y).

    Note the speed difference:
    * 14.3 µs: xp.vecdot(matrix_vector_mul(m, x), x)
    * 4.44 µs: np.einsum("...i,...ij,...j->...", x, m, x)
    """
    xp = array_namespace(x, y)
    y = xp.reshape(y, (*y.shape[:-1], 1, y.shape[-1]))
    return xp.sum(x * y, axis=-1)


def matrix_dot_product[T: Array](x: T, y: T) -> T:
    """Return the "matrix dot product" of a matrix with the outer product of a vector.

    This equals:
    * 1.19 µs: xp.einsum("...ij,...ij", x, y)
    * 1.77 µs: xp.sum(x * y, axis=(-2, -1))
    # 3.87 µs: xp.linalg.trace(xp.moveaxis(x, -2, -1) @ y)
    """
    xp = array_namespace(x, y)
    return xp.sum(x * y, axis=(-2, -1))


def divide_where[T: Array](
    dividend: T, divisor: T, *, where: BooleanArray | None = None, otherwise: T | None = None
) -> T:
    """Return the quotient or a special value when a condition is false.

    Returns: `xp.where(where, dividend / divisor, otherwise)`, but without evaluating
    `dividend / divisor` when `where` is false.  This prevents both infinite cotangents, and
    some exceptions.
    """
    if where is None:
        assert otherwise is None
        xp = array_namespace(dividend, divisor)
        return xp.divide(dividend, divisor)
    assert otherwise is not None
    xp = array_namespace(dividend, divisor, where, otherwise)
    dividend = xp.where(where, dividend, 1.0)
    divisor = xp.where(where, divisor, 1.0)
    quotient: T = xp.divide(dividend, divisor)
    return xp.where(where, quotient, otherwise)


# Remove when https://github.com/scipy/scipy/pull/18605 is released.
def softplus[T: Array](x: T, /, *, xp: Namespace | None = None) -> T:
    """Softplus, which is log(1+exp(x)).

    This has asymptotic behavior exp(x) as x -> -inf and x as x -> +inf.
    """
    if xp is None:
        xp = array_namespace(x)
    return xp.logaddexp(xp.asarray(0.0), x)


def log_softplus[T: Array](x: T, /, *, xp: Namespace | None = None) -> T:
    """Log-softplus, which is log(1+log(1+exp(x))).

    This has asymptotic behavior 0 as x -> -inf and log(x) as x -> +inf.
    """
    if xp is None:
        xp = array_namespace(x)
    z = xp.asarray(0.0)
    return xp.logaddexp(z, xp.logaddexp(z, x))


def bessel_iv_ratio(v: jax.Array, x: jax.Array, /, *, iterations: int = 200) -> jax.Array:
    """Return ``I_(v + 1)(x) / I_v(x)`` for nonnegative real ``x``.

    The implementation uses a backward recurrence, which is stable for the ratio and remains fully
    differentiable with respect to both ``v`` and ``x``.
    """
    v, x = jnp.broadcast_arrays(jnp.asarray(v), jnp.asarray(x))
    q = jnp.zeros_like(v + x)

    def body(i: int, q: jax.Array) -> jax.Array:
        order = v + (iterations - i)
        return x / (2.0 * order + x * q)

    return jax.lax.fori_loop(0, iterations, body, q)


def _scipy_log_bessel_ive(v: np.ndarray, x: np.ndarray) -> np.ndarray:
    return np.log(sc.ive(v, x))


def _log_bessel_ive_callback(v: jax.Array, x: jax.Array) -> jax.Array:
    v, x = jnp.broadcast_arrays(jnp.asarray(v), jnp.asarray(x))
    shape = v.shape
    dtype = jnp.result_type(v, x, jnp.float32)

    def callback(v_np: np.ndarray, x_np: np.ndarray) -> np.ndarray:
        return _scipy_log_bessel_ive(v_np, x_np).astype(dtype, copy=False)

    return jax.pure_callback(callback, jax.ShapeDtypeStruct(shape, dtype), v, x)


@jax.custom_jvp
def log_bessel_ive(v: jax.Array, x: jax.Array, /) -> jax.Array:
    """Return ``log(ive(v, x))`` for nonnegative real ``x``.

    The primal value is delegated to SciPy because JAX does not currently expose ``ive`` for
    general real order. The custom JVP provides derivatives with respect to both ``v`` and ``x``.
    """
    return _log_bessel_ive_callback(v, x)


@log_bessel_ive.defjvp
def _log_bessel_ive_jvp(
    primals: tuple[jax.Array, jax.Array], tangents: tuple[jax.Array, jax.Array]
) -> tuple[jax.Array, jax.Array]:
    v, x = primals
    v_dot, x_dot = tangents
    primal_out = log_bessel_ive(v, x)

    safe_x = jnp.maximum(jnp.asarray(x), jnp.sqrt(jnp.finfo(jnp.result_type(x, jnp.float32)).eps))
    x_tangent = x_dot * (bessel_iv_ratio(v, safe_x) + v / safe_x - 1.0)

    step = jnp.cbrt(jnp.finfo(jnp.result_type(v, jnp.float32)).eps) * (1.0 + jnp.abs(v))
    v_tangent = (
        v_dot
        * (_log_bessel_ive_callback(v + step, x) - _log_bessel_ive_callback(v - step, x))
        / (2.0 * step)
    )
    return primal_out, v_tangent + x_tangent


def sublinear_softplus[T: Array](x: T, maximum: T, /, *, xp: Namespace | None = None) -> T:
    """Sublinear-softplus, which is softplus(x) / (1 + softplus(x) / maximum).

    This has asymptotic behavior exp(x) as x -> -inf and maximum as x -> +inf.
    """
    if xp is None:
        xp = array_namespace(x)
    o = xp.asarray(1.0)
    sp = softplus(x, xp=xp)
    return sp / (o + sp / maximum)


def inverse_softplus[T: Array](y: T, /, *, xp: Namespace | None = None) -> T:
    if xp is None:
        xp = array_namespace(y)
    return xp.where(
        y > 80.0,  # noqa: PLR2004
        y,
        xp.log(xp.expm1(y)),
    )


def normalize[T: Array](
    mode: Literal["l1", "l2", "max"], x: T, *, axis: tuple[int, ...] | int | None = None
) -> T:
    """Returns the L1-normalized copy of x, assuming that x is nonnegative."""
    xp = array_namespace(x)
    epsilon = 10 * xp.finfo(x.dtype).eps
    match mode:
        case "l1":
            sum_x = xp.sum(xp.abs(x), axis=axis, keepdims=True)
            size = x.size / sum_x.size
            return xp.where(sum_x < epsilon, xp.ones_like(x) / size, x / sum_x)
        case "l2":
            sum_x = xp.sqrt(xp.sum(xp.square(x), axis=axis, keepdims=True))
            size = x.size / sum_x.size
            return xp.where(sum_x < epsilon, xp.ones_like(x) * xp.pow(size, -0.5), x / sum_x)
        case "max":
            sum_x = xp.max(xp.abs(x), axis=axis, keepdims=True)
            return xp.where(sum_x < epsilon, xp.ones_like(x), x / sum_x)


def stop_gradient[U](x: U, *, xp: Namespace | None = None) -> U:
    if xp is None:
        xp = array_namespace(x)
    if is_jax_namespace(xp):
        from jax.lax import stop_gradient as sg  # noqa: PLC0415

        return sg(x)
    if is_torch_namespace(xp):
        from torch import Tensor  # noqa: PLC0415 # pyright: ignore # type: ignore

        assert isinstance(x, Tensor)
        return x.detach()  # pyright: ignore
    return x



from __future__ import annotations

from collections.abc import Hashable
from dataclasses import dataclass
from typing import Any, override


@dataclass(frozen=True)
class GraphNodeKey:
    """Struct for use with :func:`jax.tree_util.register_pytree_with_keys`.

    Ordering assumes node keys are mutually comparable. Heterogeneous node-key types may therefore
    fail when an ordered traversal is requested.
    """

    key: Hashable

    @override
    def __str__(self) -> str:
        return f".node[{self.key!r}]"

    def __lt__(self, value: GraphNodeKey | GraphEdgeKey, /) -> bool:
        return not isinstance(value, GraphNodeKey) or self.key < value.key  # type: ignore # pyright: ignore


@dataclass(frozen=True)
class GraphEdgeKey:
    """Struct for use with :func:`jax.tree_util.register_pytree_with_keys`.

    Ordering assumes source and target keys are mutually comparable. Heterogeneous endpoint types
    may therefore fail when an ordered traversal is requested.
    """

    source: Hashable  # TODO: mark this as comparable
    target: Hashable

    @override
    def __str__(self) -> str:
        return f".edge[{self.source!r}, {self.target!r}]"

    def __lt__(self, value: GraphNodeKey | GraphEdgeKey, /) -> bool:
        return isinstance(value, GraphEdgeKey) and (
            (self.source, self.target) < (value.source, value.target)  # type: ignore
        )


@dataclass(frozen=True)
class UndirectedGraphEdgeKey(GraphEdgeKey):
    """An undirected graph edge key with the same comparability assumptions as GraphEdgeKey."""

    @override
    def __hash__(self) -> int:
        return hash(self.source) ^ hash(self.target)

    @override
    def __lt__(self, value: GraphNodeKey | GraphEdgeKey, /) -> bool:
        return isinstance(value, GraphEdgeKey) and (
            sorted((self.source, self.target))  # pyright: ignore
            < sorted((value.source, value.target))  # pyright: ignore
        )


type GraphAuxData = list[GraphNodeKey | GraphEdgeKey]  # The auxilliary data for Jax.
type GraphData = dict[str, Any]  # What networkx stores in the graph.



from __future__ import annotations

from typing import override

from rich.console import Console

from .print_generic import print_generic


class GenericString:
    """An object that can be passed to a logging function to lazily log using print_generic."""

    def __init__(
        self,
        *args: object,
        immediate: bool = True,
        raise_on_nan: bool = True,
        as_leaves: bool = False,
    ) -> None:
        super().__init__()
        self.args = args
        self.raise_on_nan = raise_on_nan
        self.immediate = immediate
        self.as_leaves = as_leaves

    @override
    def __str__(self) -> str:
        console = Console(no_color=True)
        with console.capture() as capture:
            print_generic(
                *self.args,
                raise_on_nan=self.raise_on_nan,
                immediate=self.immediate,
                as_leaves=self.as_leaves,
                console=console,
            )
        return capture.get()



from io import StringIO
from textwrap import dedent
from typing import Any

import jax.numpy as jnp
import jax.random as jr
import numpy as np
import pytest
from jax import Array, enable_custom_prng, jit, tree, vmap
from rich.console import Console

from tjax import KeyArray, RealArray, print_generic
from tjax.dataclasses import dataclass, field


def verify(actual: str, desired: str) -> None:
    actual = "\n".join(x.rstrip() for x in actual.strip().split("\n"))
    assert actual == dedent(desired).strip()


def test_numpy_display(capsys: pytest.CaptureFixture[str], console: Console) -> None:
    print_generic(np.reshape(np.arange(6.0), (3, 2)), console=console)
    assert isinstance(console.file, StringIO)
    captured = console.file.getvalue()
    verify(
        captured,
        """
           NumPy Array (3, 2) float64
           └──  0.0000 │ 1.0000
                2.0000 │ 3.0000
                4.0000 │ 5.0000
           """,
    )


def test_numpy_display_big(capsys: pytest.CaptureFixture[str], console: Console) -> None:
    print_generic(np.reshape(np.arange(30.0), (15, 2)), console=console)
    assert isinstance(console.file, StringIO)
    captured = console.file.getvalue()
    verify(
        captured,
        """
           NumPy Array (15, 2) float64
           ├── mean=14.5
           └── deviation=8.65544144839919
           """,
    )


def test_jax_numpy_display(capsys: pytest.CaptureFixture[str], console: Console) -> None:
    print_generic(jnp.ones(3), console=console)
    assert isinstance(console.file, StringIO)
    captured = console.file.getvalue()
    verify(
        captured,
        """
           Jax Array (3,) float64
           └──  1.0000 │ 1.0000 │ 1.0000
           """,
    )


def test_jit_display(capsys: pytest.CaptureFixture[str], console: Console) -> None:
    @jit
    def f(x: RealArray) -> RealArray:
        print_generic(x, console=console)
        return x

    f(jnp.asarray(1.0))
    assert isinstance(console.file, StringIO)
    captured = console.file.getvalue()
    verify(
        captured,
        """
           Jax Array () float64
           └── 1.0000
           """,
    )


def test_vmap_display(capsys: pytest.CaptureFixture[str], console: Console) -> None:
    """Like test_batch_display, but uses print_generic to get the array."""

    @jit
    def f(x: RealArray) -> RealArray:
        print_generic(x, console=console)
        return x

    vmap(vmap(f, in_axes=2), in_axes=1)(jnp.ones((3, 4, 5, 6)))
    assert isinstance(console.file, StringIO)
    captured = console.file.getvalue()
    s = dedent("""
               Jax Array (3, 5) float64
               └──  1.0000 │ 1.0000 │ 1.0000 │ 1.0000 │ 1.0000
                    1.0000 │ 1.0000 │ 1.0000 │ 1.0000 │ 1.0000
                    1.0000 │ 1.0000 │ 1.0000 │ 1.0000 │ 1.0000""")
    verify(captured, s * 24)


def test_immediate_key_display(capsys: pytest.CaptureFixture[str], console: Console) -> None:
    with enable_custom_prng():
        k = jr.key(123)

        @jit
        def f(x: KeyArray) -> None:
            print_generic(x, immediate=True, console=console)

        f(k)
    assert isinstance(console.file, StringIO)
    captured = console.file.getvalue()
    verify(
        captured,
        """
           JaxKey[dataclass]
           └── key_data=Jax Array (2,) uint32
           """,
    )


def test_key_display(capsys: pytest.CaptureFixture[str], console: Console) -> None:
    with enable_custom_prng():
        k = jr.key(123)

        @jit
        def f(x: KeyArray) -> None:
            print_generic(x, console=console)

        f(k)
    assert isinstance(console.file, StringIO)
    captured = console.file.getvalue()
    verify(
        captured,
        """
           JaxKey[dataclass]
           └── key_data=Jax Array (2,) uint32
               └──  0 │ 123
           """,
    )


def test_dict_display(capsys: pytest.CaptureFixture[str], console: Console) -> None:
    print_generic({"cat": 5, "mouse": {"dog": 3, "sheep": 4}}, console=console)
    assert isinstance(console.file, StringIO)
    captured = console.file.getvalue()
    verify(
        captured,
        """
           dict
           ├── cat=5
           └── mouse=dict
               ├── dog=3
               └── sheep=4
           """,
    )


def test_dataclass(capsys: pytest.CaptureFixture[str], console: Console) -> None:
    @dataclass
    class C:
        x: int
        y: int

    @dataclass
    class D:
        c: C
        d: C

    print_generic(D(C(1, 2), C(3, 4)), console=console)
    assert isinstance(console.file, StringIO)
    captured = console.file.getvalue()
    verify(
        captured,
        """
           D[dataclass]
           ├── c=C[dataclass]
           │   ├── x=1
           │   └── y=2
           └── d=C[dataclass]
               ├── x=3
               └── y=4
           """,
    )


def test_dataclass_as_leaves(capsys: pytest.CaptureFixture[str], console: Console) -> None:
    @dataclass
    class C:
        x: int
        y: int

    @dataclass
    class D:
        c: C
        d: C

    print_generic(D(C(1, 2), C(3, 4)), console=console, as_leaves=True)
    assert isinstance(console.file, StringIO)
    captured = console.file.getvalue()
    verify(
        captured,
        """
           list
           ├── tuple
           │   ├── ".c.x"
           │   └── 1
           ├── tuple
           │   ├── ".c.y"
           │   └── 2
           ├── tuple
           │   ├── ".d.x"
           │   └── 3
           └── tuple
               ├── ".d.y"
               └── 4
           """,
    )


def test_pytreedef(capsys: pytest.CaptureFixture[str], console: Console) -> None:
    @dataclass
    class C:
        x: int
        y: int

    _, tree_def = tree.flatten(C(1, 2))

    print_generic(tree_def, immediate=True, console=console)
    assert isinstance(console.file, StringIO)
    captured = console.file.getvalue()
    verify(
        captured,
        f"""
           PyTreeDef
           └── hash={hash(tree_def)}
           """,
    )


def test_seen_array(capsys: pytest.CaptureFixture[str], console: Console) -> None:
    @dataclass
    class C:
        x: Array
        y: Array

    z = jnp.zeros(2)
    tree_def = C(z, z)

    print_generic(tree_def, immediate=True, console=console)
    assert isinstance(console.file, StringIO)
    captured = console.file.getvalue()
    verify(
        captured,
        """
           C[dataclass]
           ├── x=Jax Array (2,) float64
           │   └──  0.0000 │ 0.0000
           └── y=Jax Array (2,) float64
               └──  0.0000 │ 0.0000
           """,
    )


if __name__ == "__main__":

    @dataclass
    class Triplet:
        x: Any
        y: Any
        z: Any = field(static=True)

    a = Triplet(np.reshape(np.arange(6.0), (3, 2)), np.reshape(np.arange(30.0), (15, 2)), Triplet)

    @jit
    def g(x: RealArray) -> None:
        print_generic(Triplet({"abc": Triplet(a, x, 2)}, a, "blah"))
        print_generic(Triplet({"abc": Triplet(a, x, 2)}, a, "blah"))

    vmap(vmap(g, in_axes=2), in_axes=1)(jnp.ones((3, 4, 5, 6)))



from __future__ import annotations

from functools import WRAPPER_ASSIGNMENTS

abstract_jit_marker = "_abstract_jit"
abstract_custom_jvp_marker = "_abstract_custom_jvp"
all_wrapper_assignments = (
    *WRAPPER_ASSIGNMENTS,
    "__isabstractmethod__",
    "__override__",
    abstract_jit_marker,
    abstract_custom_jvp_marker,
)
